#!/usr/bin/env python3

import unittest
from twoping import monotonic_clock


clock = monotonic_clock.clock
clock_info = monotonic_clock.get_clock_info('clock')


class TestMonotonicClock(unittest.TestCase):
    def test_clock(self):
        self.assertEqual(type(clock()), float)

    @unittest.skipUnless(clock_info.monotonic, 'Monotonic clock required')
    def test_monotonic(self):
        time1 = clock()
        time2 = clock()
        self.assertGreaterEqual(time2, time1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import unittest
from twoping import utils


class TestUtils(unittest.TestCase):
    def test_twoping_checksum_hello(self):
        data = b'Hello World'
        self.assertEqual(utils.twoping_checksum(data), 0xae31)

    def test_twoping_checksum_junk(self):
        data = b'E\xc6\xca\x92\x10\x0e\xb1\xcf\x98\x88\x0bB\xc0\xf8X\xac\xd9\x81v\xc0E\x8co\x04\x9c\x0e\x93\xb3' + \
            b'I=8m\xb7\xd5\x86\xa7f,2\x15\xd9\x7f\xad>\x1a\xe09\x89J\xdd\x0b(\xa0\x07aGT\xc6P\x04-(NH<@\xc7\xdb' + \
            b'\x82\xfa\x0b\xab7n#^\x80\xab\xb9\xe5\x05\x14\xb1\xfcrT7\x8af\xf2Q\xa1w\xda\xd5\x88\xd58C10\xe7' + \
            b'\x1ez+\xc8\x14\rK|3\x9a\x9b\xc1\xc1\xc8\xc5\xfeZHD\x1b\x87\x9d$\xd9"}'
        self.assertEqual(utils.twoping_checksum(data), 0x4a06)

    def test_twoping_checksum_packet(self):
        data = bytearray(
            b'\x32\x50\x8d\x3b\x00\x00\x00\x00\xb0\x06\x00\x3b\x00\x00\x00\x06\x00\x00\x00\x00\xa0\x0a\x00' +
            b'\x08\x00\x01\x00\x00\x00\x00\xa0\x01\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x02\x00\x08\x00\x01' +
            b'\x00\x00\x00\x00\xb0\x02'
        )
        data[2] = 0
        data[3] = 0
        self.assertEqual(utils.twoping_checksum(data), 0x8d3b)

    def test_twoping_checksum_iter(self):
        i = 65535
        for x in range(256):
            for y in range(256):
                data = bytes([x, y])
                checksum = utils.twoping_checksum(data)
                self.assertEqual(checksum, i)
                i = i - 1
                if i == 0:
                    i = 65535


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import random
import socket
import math
import signal
import sys
import errno
import time
from . import __version__
from . import packets
from . import monotonic_clock
from . import best_poller
from .args import parse_args
from .utils import _, _pl, lazy_div, npack, nunpack, platform_info, twoping_checksum

try:
    random_sys = random.SystemRandom()
    has_sysrandom = True
except AttributeError:
    random_sys = random
    has_sysrandom = False

try:
    import dns.resolver
    has_dns = True
except ImportError:
    has_dns = False

try:
    import netifaces
    has_netifaces = True
except ImportError:
    has_netifaces = False

assert(sys.version_info > (3, 4))

version_string = '2ping {} - {}'.format(__version__, platform_info())
clock = monotonic_clock.clock
clock_info = monotonic_clock.get_clock_info('clock')


class SocketClass():
    def __init__(self, sock):
        self.sock = sock

        # Used during client mode for the host tuple to send UDP packets to.
        self.client_host = None

        # Dict of PeerState instances, indexed by peer tuple
        self.peer_states = {}

        # Statistics
        self.pings_transmitted = 0
        self.pings_received = 0
        self.packets_transmitted = 0
        self.packets_received = 0
        self.lost_outbound = 0
        self.lost_inbound = 0
        self.errors_received = 0
        self.rtt_total = 0
        self.rtt_total_sq = 0
        self.rtt_count = 0
        self.rtt_min = 0
        self.rtt_max = 0
        self.rtt_ewma = 0

        self.next_send = 0

        self.shutdown_time = 0
        self.nagios_result = 0
        self.session = b''

    def fileno(self):
        return self.sock.fileno()


class PeerState():
    def __init__(self, peer_tuple, sock_class):
        self.peer_tuple = peer_tuple
        self.sock_class = sock_class

        # In-flight outbound messages.  Added in the following conditions:
        #   * Outbound packet with OpcodeReplyRequested sent.
        # Removed in following conditions:
        #   * Inbound packet with OpcodeInReplyTo set to it.
        #   * Inbound packet with it in OpcodeInvestigationSeen or OpcodeInvestigationUnseen.
        #   * Cleanup after 10 minutes.
        # If it remains for more than <10> seconds, it it sent as part of
        # OpcodeInvestigate with the next outbound packet with OpcodeReplyRequested set.
        self.sent_messages = {}
        # Seen inbound messages.  Added in the following conditions:
        #   * Inbound packet with OpcodeReplyRequested set.
        # Referenced in the following conditions:
        #   * Inbound packet with it in OpcodeInvestigate.
        # Removed in the following conditions:
        #   * Inbound packet with it in OpcodeCourtesyExpiration.
        #   * Cleanup after 10 minutes.
        self.seen_messages = {}
        # Courtesy messages waiting to be sent.  Added in the following conditions:
        #   * Inbound packet with OpcodeInReplyTo set.
        # Removed in the following conditions:
        #   * Outbound packet where there is room to send it as part of OpcodeCourtesyExpiration.
        #   * Cleanup after 2 minutes.
        self.courtesy_messages = {}
        # Current position of a peer tuple's incrementing ping integer.
        self.ping_position = 0
        # Current encrypted session ID
        self.encrypted_session_id = None
        # Seen encrypted session IVs
        self.encrypted_session_ivs = {}
        # Last time a peer was sent to or received from
        self.last_seen = clock()


class TwoPing():
    def __init__(self, args):
        now = clock()
        self.args = args
        self.time_start = now
        self.fake_time_epoch = random_sys.random() * (2**32)
        self.fake_time_generation = random_sys.randint(0, 65535)

        self.sock_classes = []
        self.poller = best_poller.best_poller()

        self.pings_transmitted = 0
        self.pings_received = 0
        self.packets_transmitted = 0
        self.packets_received = 0
        self.lost_outbound = 0
        self.lost_inbound = 0
        self.errors_received = 0
        self.rtt_total = 0
        self.rtt_total_sq = 0
        self.rtt_count = 0
        self.rtt_min = 0
        self.rtt_max = 0
        self.rtt_ewma = 0

        # Scheduled events
        self.next_cleanup = now + 60.0
        self.next_stats = 0
        if self.args.stats:
            self.next_stats = now + self.args.stats

        self.old_age_interval = 60.0
        # On Windows, KeyboardInterrupt during select() will not be trapped until a socket event or timeout, so we should set
        # the timeout to a short value.
        if sys.platform.startswith(('win32', 'cygwin')):
            self.old_age_interval = 1.0

        # Test for IPv6 functionality.
        self.has_ipv6 = True
        if not socket.has_ipv6:
            self.has_ipv6 = False
        else:
            # BSD jails seem to have has_ipv6 = True, but will throw "Protocol not supported" on bind.  Test for this.
            try:
                socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
            except socket.error as e:
                if e.errno == errno.EPROTONOSUPPORT:
                    self.has_ipv6 = False
                else:
                    raise

        self.is_reload = False

        if self.args.send_monotonic_clock and (not clock_info.monotonic):
            self.args.send_monotonic_clock = False

    def print_out(self, *args, **kwargs):
        print(*args, **kwargs)

    def print_debug(self, *args, **kwargs):
        if not self.args.debug:
            return
        self.print_out(*args, **kwargs)

    def shutdown(self):
        self.print_stats()
        if self.args.nagios:
            sys.exit(self.nagios_result)
        sys.exit(0)

    def handle_socket_error(self, e, sock_class, peer_address=None):
        sock = sock_class.sock
        # Errors from the last send() can be trapped via IP_RECVERR (Linux only).
        self.errors_received += 1
        sock_class.errors_received += 1
        error_string = str(e)
        try:
            MSG_ERRQUEUE = 8192
            (error_data, error_address) = sock.recvfrom(16384, MSG_ERRQUEUE)
            if self.args.quiet:
                pass
            elif self.args.flood:
                self.print_out('E', end='', flush=True)
            else:
                self.print_out('{}: {}'.format(error_address[0], error_string))
        except socket.error:
            if self.args.quiet:
                pass
            elif self.args.flood:
                self.print_out('E', end='', flush=True)
            else:
                if peer_address:
                    self.print_out('{}: {}'.format(peer_address[0], error_string))
                else:
                    self.print_out(error_string)

    def process_incoming_packet(self, sock_class):
        sock = sock_class.sock
        try:
            (data, peer_address) = sock.recvfrom(16384)
        except socket.error as e:
            self.handle_socket_error(e, sock_class)
            return
        socket_address = sock.getsockname()
        self.print_debug('Socket address: {}'.format(repr(socket_address)))
        self.print_debug('Peer address: {}'.format(repr(peer_address)))

        # Simulate random packet loss.
        if self.args.packet_loss_in and (random.random() < (self.args.packet_loss_in / 100.0)):
            return
        # Simulate data corruption
        if self.args.fuzz:
            data = self.fuzz_packet(data)

        # Per-packet options.
        self.packets_received += 1
        sock_class.packets_received += 1
        calculated_rtt = None

        time_begin = clock()
        peer_tuple = (socket_address, peer_address, sock.type)

        # Preload state tables if the client has not been seen (or has been cleaned).
        if peer_tuple not in sock_class.peer_states:
            sock_class.peer_states[peer_tuple] = PeerState(peer_tuple, sock_class)
        peer_state = sock_class.peer_states[peer_tuple]
        peer_state.last_seen = time_begin

        # Load/parse the packet.
        packet_in = packets.Packet()
        packet_in.load(data)
        if self.args.verbose:
            self.print_out('RECV: {}'.format(repr(packet_in)))

        # Decrypt packet if needed.
        if self.args.encrypt:
            if packets.OpcodeEncrypted.id not in packet_in.opcodes:
                self.errors_received += 1
                sock_class.errors_received += 1
                self.print_out(_('Encryption required but not provided by {address}').format(address=peer_address[0]))
                return
            if packet_in.opcodes[packets.OpcodeEncrypted.id].method_index != self.args.encrypt_method_index:
                self.errors_received += 1
                sock_class.errors_received += 1
                self.print_out(
                    _('Encryption method mismatch from {address} (expected {expected}, got {got})').format(
                        address=peer_address[0],
                        expected=self.args.encrypt_method_index,
                        got=packet_in.opcodes[packets.OpcodeEncrypted.id].method_index,
                    )
                )
                return
            data = packet_in.opcodes[packets.OpcodeEncrypted.id].decrypt(self.args.encrypt.encode('UTF-8'))
            encrypted_packet_in = packet_in
            packet_in = packets.Packet()
            packet_in.load(data)
            if peer_state.encrypted_session_id is None:
                peer_state.encrypted_session_id = encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].session
            if encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].session != peer_state.encrypted_session_id:
                self.errors_received += 1
                sock_class.errors_received += 1
                self.print_out(
                    _('Encryption session mismatch from {address} (expected {expected}, got {got})').format(
                        address=peer_address[0],
                        expected=repr(peer_state.encrypted_session_id),
                        got=repr(encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].session),
                    )
                )
                return
            if encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].iv in peer_state.encrypted_session_ivs:
                self.errors_received += 1
                sock_class.errors_received += 1
                self.print_out(
                    _('Repeated IV {iv} from {address}, discarding').format(
                        iv=repr(encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].iv),
                        address=peer_address[0],
                    )
                )
                return
            peer_state.encrypted_session_ivs[encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].iv] = (time_begin,)
            if self.args.verbose:
                self.print_out('DECR: {}'.format(repr(packet_in)))

        # Verify HMAC if required.
        if self.args.auth:
            if packets.OpcodeHMAC.id not in packet_in.opcodes:
                self.errors_received += 1
                sock_class.errors_received += 1
                self.print_out(_('Auth required but not provided by {address}').format(address=peer_address[0]))
                return
            if packet_in.opcodes[packets.OpcodeHMAC.id].digest_index != self.args.auth_digest_index:
                self.errors_received += 1
                sock_class.errors_received += 1
                self.print_out(
                    _('Auth digest type mismatch from {address} (expected {expected}, got {got})').format(
                        address=peer_address[0],
                        expected=self.args.auth_digest_index,
                        got=packet_in.opcodes[packets.OpcodeHMAC.id].digest_index,
                    )
                )
                return
            (test_begin, test_length) = packet_in.opcode_data_positions[packets.OpcodeHMAC.id]
            test_begin += 2
            test_length -= 2
            packet_in.opcodes[packets.OpcodeHMAC.id].key = self.args.auth.encode('UTF-8')
            test_data = bytearray(data)
            test_data[2:4] = b'\x00\x00'
            test_data[test_begin:(test_begin+test_length)] = bytes(test_length)
            test_hash = packet_in.opcodes[packets.OpcodeHMAC.id].hash
            test_hash_calculated = packet_in.calculate_hash(packet_in.opcodes[packets.OpcodeHMAC.id], test_data)
            if test_hash_calculated != test_hash:
                self.errors_received += 1
                sock_class.errors_received += 1
                self.print_out(
                    _('Auth hash failed from {address} (expected {expected}, got {got})').format(
                        address=peer_address[0],
                        expected=''.join('{hex:02x}'.format(hex=x) for x in test_hash_calculated),
                        got=''.join('{hex:02x}'.format(hex=x) for x in test_hash),
                    )
                )
                return

        # If this is in reply to one of our sent packets, it's a ping reply, so handle it specially.
        if packets.OpcodeInReplyTo.id in packet_in.opcodes:
            replied_message_id = packet_in.opcodes[packets.OpcodeInReplyTo.id].message_id
            replied_message_id_int = nunpack(replied_message_id)
            if replied_message_id_int in peer_state.sent_messages:
                (sent_time, _unused, ping_position) = peer_state.sent_messages[replied_message_id_int]
                del(peer_state.sent_messages[replied_message_id_int])
                calculated_rtt = (time_begin - sent_time) * 1000
                self.pings_received += 1
                sock_class.pings_received += 1
                self.update_rtts(sock_class, calculated_rtt)
                if self.args.quiet:
                    pass
                elif self.args.flood:
                    self.print_out('\x08', end='', flush=True)
                else:
                    if self.args.audible:
                        self.print_out('\x07', end='', flush=True)
                    if packets.OpcodeRTTEnclosed.id in packet_in.opcodes:
                        self.print_out(
                            _('{bytes} bytes from {address}: ping_seq={seq} time={ms:0.03f} ms peertime={peerms:0.03f} ms').format(
                                bytes=len(data),
                                address=peer_state.peer_tuple[1][0],
                                seq=ping_position,
                                ms=calculated_rtt,
                                peerms=(packet_in.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us / 1000.0),
                            )
                        )
                    else:
                        self.print_out(
                            _('{bytes} bytes from {address}: ping_seq={seq} time={ms:0.03f} ms').format(
                                bytes=len(data),
                                address=peer_state.peer_tuple[1][0],
                                seq=ping_position,
                                ms=calculated_rtt,
                            )
                        )
                    if (
                        (packets.OpcodeExtended.id in packet_in.opcodes) and
                        (packets.ExtendedNotice.id in packet_in.opcodes[packets.OpcodeExtended.id].segments)
                    ):
                        notice = packet_in.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedNotice.id].text
                        self.print_out('  ' + _('Peer notice: {notice}').format(notice=notice))
            peer_state.courtesy_messages[replied_message_id_int] = (time_begin, replied_message_id)

        # Check if any invesitgations results have come back.
        self.check_investigations(peer_state, packet_in)

        # Process courtesy expirations
        if packets.OpcodeCourtesyExpiration.id in packet_in.opcodes:
            for message_id in packet_in.opcodes[packets.OpcodeCourtesyExpiration.id].message_ids:
                message_id_int = nunpack(message_id)
                if message_id_int in peer_state.seen_messages:
                    del(peer_state.seen_messages[message_id_int])

        # If the peer requested a reply, prepare one.
        if packets.OpcodeReplyRequested.id in packet_in.opcodes:
            # Populate seen_messages.
            peer_state.seen_messages[nunpack(packet_in.message_id)] = (time_begin,)

            # Basic packet configuration.
            packet_out = self.base_packet()
            packet_out.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
            packet_out.opcodes[packets.OpcodeInReplyTo.id].message_id = packet_in.message_id

            # If we are matching packet sizes of the peer, adjust the minimum if it falls between min_packet_size
            # and max_packet_size.
            if not self.args.no_match_packet_size:
                data_len = len(data)
                if (data_len <= self.args.max_packet_size) and (data_len >= self.args.min_packet_size):
                    packet_out.min_length = data_len

            # 3-way pings already have the first roundtrip calculated.
            if calculated_rtt is not None:
                packet_out.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
                packet_out.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = int(calculated_rtt * 1000)

            # Check for any investigations the peer requested.
            if packets.OpcodeInvestigate.id in packet_in.opcodes:
                for message_id in packet_in.opcodes[packets.OpcodeInvestigate.id].message_ids:
                    if nunpack(message_id) in peer_state.seen_messages:
                        if packets.OpcodeInvestigationSeen.id not in packet_out.opcodes:
                            packet_out.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
                        packet_out.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(message_id)
                    else:
                        if packets.OpcodeInvestigationUnseen.id not in packet_out.opcodes:
                            packet_out.opcodes[packets.OpcodeInvestigationUnseen.id] = packets.OpcodeInvestigationUnseen()
                        packet_out.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids.append(message_id)

            # If the packet_in is ReplyRequested but not InReplyTo, it is a second leg.  Unless 3-way ping was
            # disabled, request a reply.
            if (packets.OpcodeInReplyTo.id not in packet_in.opcodes) and (not self.args.no_3way):
                packet_out.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()

            # Send any investigations we would like to know about.
            self.start_investigations(peer_state, packet_out)

            # Any courtesy expirations we have waiting should be sent.
            if len(peer_state.courtesy_messages) > 0:
                packet_out.opcodes[packets.OpcodeCourtesyExpiration.id] = packets.OpcodeCourtesyExpiration()
                for (courtesy_time, courtesy_message_id) in peer_state.courtesy_messages.values():
                    packet_out.opcodes[packets.OpcodeCourtesyExpiration.id].message_ids.append(courtesy_message_id)

            # Calculate the host latency as late as possible.
            packet_out.opcodes[packets.OpcodeHostLatency.id] = packets.OpcodeHostLatency()
            time_send = clock()
            packet_out.opcodes[packets.OpcodeHostLatency.id].delay_us = int((time_send - time_begin) * 1000000)

            # Dump the packet.
            dump_out = packet_out.dump()

            # If enabled, encrypt the packet and wrap it in a stub packet.
            if self.args.encrypt:
                encrypted_packet = packets.Packet()
                encrypted_packet.opcodes[packets.OpcodeEncrypted.id] = packets.OpcodeEncrypted()
                encrypted_packet.opcodes[packets.OpcodeEncrypted.id].method_index = self.args.encrypt_method_index
                encrypted_packet.opcodes[packets.OpcodeEncrypted.id].session = encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].session
                encrypted_packet.opcodes[packets.OpcodeEncrypted.id].encrypt(dump_out, self.args.encrypt.encode('UTF-8'))
                sock_out = encrypted_packet.dump()
            else:
                sock_out = dump_out

            # Send the packet.
            self.sock_sendto(sock_class, sock_out, peer_address)
            self.packets_transmitted += 1
            sock_class.packets_transmitted += 1

            # If ReplyRequested is set, we care about its arrival.
            if packets.OpcodeReplyRequested.id in packet_out.opcodes:
                self.pings_transmitted += 1
                sock_class.pings_transmitted += 1
                peer_state.ping_position += 1
                peer_state.sent_messages[nunpack(packet_out.message_id)] = (
                    time_send,
                    packet_out.message_id,
                    peer_state.ping_position
                )

            # Examine the sent packet.
            packet_out_examine = packets.Packet()
            packet_out_examine.load(dump_out)

            # Any courtesy expirations which had room in the sent packet should be forgotten.
            if packets.OpcodeCourtesyExpiration.id in packet_out_examine.opcodes:
                for courtesy_message_id in packet_out_examine.opcodes[packets.OpcodeCourtesyExpiration.id].message_ids:
                    courtesy_message_id_int = nunpack(courtesy_message_id)
                    if courtesy_message_id_int in peer_state.courtesy_messages:
                        del(peer_state.courtesy_messages[courtesy_message_id_int])

            if self.args.verbose:
                if self.args.encrypt:
                    self.print_out('SEND (encrypted): {}'.format(repr(packet_out_examine)))
                else:
                    self.print_out('SEND: {}'.format(repr(packet_out_examine)))

        if (not self.args.listen) and (packets.OpcodeInReplyTo.id in packet_in.opcodes):
            if self.args.flood:
                # If we're in flood mode and this is a ping reply, send a new ping ASAP.
                sock_class.next_send = time_begin
            elif self.args.adaptive and sock_class.rtt_ewma:
                # Adaptive gets recalculated immediately after the reply.
                next_send = time_begin + (sock_class.rtt_ewma / 8.0 / 1000.0)
                if next_send < sock_class.next_send:
                    sock_class.next_send = next_send

    def sock_sendto(self, sock_class, data, address):
        sock = sock_class.sock
        # Simulate random packet loss.
        if self.args.packet_loss_out and (random.random() < (self.args.packet_loss_out / 100.0)):
            return
        # Send the packet.
        try:
            sock.sendto(data, address)
        except socket.error as e:
            self.handle_socket_error(e, sock_class, peer_address=address)

    def start_investigations(self, peer_state, packet_check):
        if len(peer_state.sent_messages) == 0:
            return
        if packets.OpcodeInvestigate.id in packet_check.opcodes:
            iobj = packet_check.opcodes[packets.OpcodeInvestigate.id]
        else:
            iobj = None
        now = clock()
        for message_id_str in peer_state.sent_messages:
            (sent_time, message_id, _unused) = peer_state.sent_messages[message_id_str]
            if now >= (sent_time + self.args.inquire_wait):
                if iobj is None:
                    iobj = packets.OpcodeInvestigate()
                if message_id not in iobj.message_ids:
                    iobj.message_ids.append(message_id)
        if iobj is not None:
            packet_check.opcodes[packets.OpcodeInvestigate.id] = iobj

    def check_investigations(self, peer_state, packet_check):
        found = {}

        # Inbound
        if packets.OpcodeInvestigationSeen.id in packet_check.opcodes:
            for message_id in packet_check.opcodes[packets.OpcodeInvestigationSeen.id].message_ids:
                message_id_int = nunpack(message_id)
                if message_id_int not in peer_state.sent_messages:
                    continue
                (_unused, _unused, ping_seq) = peer_state.sent_messages[message_id_int]
                found[ping_seq] = ('inbound', peer_state.peer_tuple[1][0])
                del(peer_state.sent_messages[message_id_int])
                self.lost_inbound += 1
                peer_state.sock_class.lost_inbound += 1

        # Outbound
        if packets.OpcodeInvestigationUnseen.id in packet_check.opcodes:
            for message_id in packet_check.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids:
                message_id_int = nunpack(message_id)
                if message_id_int not in peer_state.sent_messages:
                    continue
                (_unused, _unused, ping_seq) = peer_state.sent_messages[message_id_int]
                found[ping_seq] = ('outbound', peer_state.peer_tuple[1][0])
                del(peer_state.sent_messages[message_id_int])
                self.lost_outbound += 1
                peer_state.sock_class.lost_outbound += 1

        if self.args.quiet:
            return
        # Print results
        for ping_seq in sorted(found):
            (loss_type, address) = found[ping_seq]
            if loss_type == 'inbound':
                if self.args.flood:
                    self.print_out('<', end='', flush=True)
                else:
                    self.print_out(_('Lost inbound packet from {address}: ping_seq={seq}').format(
                        address=address,
                        seq=ping_seq,
                    ))
            else:
                if self.args.flood:
                    self.print_out('>', end='', flush=True)
                else:
                    self.print_out(_('Lost outbound packet to {address}: ping_seq={seq}').format(
                        address=address,
                        seq=ping_seq,
                    ))

    def setup_listener(self):
        # If called idempotently, destroy all listeners first
        for sock_class in self.sock_classes:
            self.poller.unregister(sock_class)
            sock_class.sock.close()

        self.sock_classes = []
        bound_addresses = []
        if self.args.all_interfaces:
            if not has_netifaces:
                raise socket.error('All interface addresses not available; please install netifaces')
            addrs = set()
            for iface in netifaces.interfaces():
                iface_addrs = netifaces.ifaddresses(iface)
                if (self.args.ipv4 or (not self.args.ipv4 and not self.args.ipv6)) and (netifaces.AF_INET in iface_addrs):
                    addrs.update([f['addr'] for f in iface_addrs[netifaces.AF_INET] if 'addr' in f])
                if self.has_ipv6 and (self.args.ipv6 or (not self.args.ipv4 and not self.args.ipv6)) and (netifaces.AF_INET6 in iface_addrs):
                    addrs.update([f['addr'] for f in iface_addrs[netifaces.AF_INET6] if 'addr' in f])
            interface_addresses = list(addrs)
        elif self.args.interface_address:
            interface_addresses = self.args.interface_address
        else:
            interface_addresses = ['0.0.0.0']
            if self.has_ipv6:
                interface_addresses.append('::')
        for interface_address in interface_addresses:
            for l in socket.getaddrinfo(
                interface_address,
                self.args.port,
                socket.AF_UNSPEC,
                socket.SOCK_DGRAM,
                socket.IPPROTO_UDP
            ):
                if l in bound_addresses:
                    continue
                if (l[0] == socket.AF_INET6) and (not self.args.ipv4) and self.has_ipv6:
                    pass
                elif (l[0] == socket.AF_INET) and (not self.args.ipv6):
                    pass
                else:
                    continue
                sock = self.new_socket(l[0], l[1], l[4])
                sock_class = SocketClass(sock)
                self.sock_classes.append(sock_class)
                self.poller.register(sock_class)
                bound_addresses.append(l)
                self.print_out(_('2PING listener ({address}): {min} to {max} bytes of data.').format(
                    address=l[4][0],
                    min=self.args.min_packet_size,
                    max=self.args.max_packet_size,
                ))

    def setup_client(self):
        if self.args.srv:
            if not has_dns:
                raise socket.error('DNS SRV lookups not available; please install dnspython')
            hosts = []
            for lookup in self.args.host:
                lookup_hosts_found = 0
                self.print_debug('SRV lookup: {}'.format(lookup))
                try:
                    res = dns.resolver.query('_{}._udp.{}'.format(self.args.srv_service, lookup), 'srv')
                except dns.exception.DNSException as e:
                    raise socket.error('{}: {}'.format(lookup, repr(e)))
                for rdata in res:
                    self.print_debug('SRV result for {}: {}'.format(
                        lookup,
                        repr(rdata),
                    ))
                    if (str(rdata.target), rdata.port) in hosts:
                        continue
                    hosts.append((str(rdata.target), rdata.port))
                    lookup_hosts_found += 1
                if lookup_hosts_found == 0:
                    raise socket.error('{}: No SRV results'.format(lookup))
        else:
            hosts = [(x, self.args.port) for x in self.args.host]
        for (hostname, port) in hosts:
            try:
                self.setup_client_host(hostname, port)
            except socket.error as e:
                eargs = list(e.args)
                if len(eargs) == 1:
                    eargs[0] = '{}: {}'.format(hostname, eargs[0])
                else:
                    eargs[1] = '{}: {}'.format(hostname, eargs[1])
                raise socket.error(*eargs)

    def setup_client_host(self, hostname, port):
        host_info = None
        for l in socket.getaddrinfo(
                hostname,
                port,
                socket.AF_UNSPEC,
                socket.SOCK_DGRAM,
                socket.IPPROTO_UDP,
                socket.AI_CANONNAME,
        ):
            if (l[0] == socket.AF_INET6) and (not self.args.ipv4) and self.has_ipv6:
                host_info = l
                break
            elif (l[0] == socket.AF_INET) and (not self.args.ipv6):
                host_info = l
                break
            else:
                continue
        if host_info is None:
            raise socket.error('Name or service not known')

        bind_info = None
        if self.args.interface_address:
            h = self.args.interface_address[-1]
        else:
            if host_info[0] == socket.AF_INET6:
                h = '::'
            else:
                h = '0.0.0.0'
        for l in socket.getaddrinfo(h, 0, host_info[0], socket.SOCK_DGRAM, socket.IPPROTO_UDP):
            bind_info = l
            break
        if bind_info is None:
            raise socket.error(_('Cannot find suitable bind for {address}').format(address=host_info[4]))
        sock = self.new_socket(bind_info[0], bind_info[1], bind_info[4])
        sock_class = SocketClass(sock)
        sock_class.client_host = host_info
        sock_class.session = bytes([random_sys.randint(0, 255) for x in range(8)])
        self.sock_classes.append(sock_class)
        self.poller.register(sock_class)
        if not self.args.nagios:
            self.print_out(
                _('2PING {hostname} ({address}): {min} to {max} bytes of data.').format(
                    hostname=host_info[3],
                    address=host_info[4][0],
                    min=self.args.min_packet_size,
                    max=self.args.max_packet_size,
                )
            )

    def fuzz_packet(self, packet):
        def fuzz_bytearray(data, pct):
            for p in range(len(data)):
                xor_byte = 0
                for i in range(8):
                    if random.random() < pct:
                        xor_byte = xor_byte + (2 ** i)
                data[p] = data[p] ^ xor_byte
            return data

        packet = bytearray(packet)
        fuzz_fraction = self.args.fuzz / 100.0

        # Fuzz the entire packet
        packet[4:] = fuzz_bytearray(packet[4:], fuzz_fraction)

        # Fuzz the magic number, at a lower probability
        packet[0:2] = fuzz_bytearray(packet[0:2], fuzz_fraction / 10.0)

        # Fuzz the recalculated checksum itself, at a lower probability
        packet[2:4] = b'\x00\x00'
        packet[2:4] = fuzz_bytearray(bytearray(npack(twoping_checksum(packet), 2)), fuzz_fraction / 10.0)

        return bytes(packet)

    def send_new_ping(self, sock_class, peer_address):
        sock = sock_class.sock
        socket_address = sock.getsockname()
        peer_tuple = (socket_address, peer_address, sock.type)
        if peer_tuple not in sock_class.peer_states:
            sock_class.peer_states[peer_tuple] = PeerState(peer_tuple, sock_class)
        peer_state = sock_class.peer_states[peer_tuple]
        peer_state.last_seen = clock()

        packet_out = self.base_packet()
        packet_out.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.start_investigations(peer_state, packet_out)
        dump_out = packet_out.dump()

        # If enabled, encrypt the packet and wrap it in a stub packet.
        if self.args.encrypt:
            encrypted_packet = packets.Packet()
            encrypted_packet.opcodes[packets.OpcodeEncrypted.id] = packets.OpcodeEncrypted()
            encrypted_packet.opcodes[packets.OpcodeEncrypted.id].method_index = self.args.encrypt_method_index
            encrypted_packet.opcodes[packets.OpcodeEncrypted.id].session = sock_class.session
            encrypted_packet.opcodes[packets.OpcodeEncrypted.id].encrypt(dump_out, self.args.encrypt.encode('UTF-8'))
            sock_out = encrypted_packet.dump()
        else:
            sock_out = dump_out

        now = clock()
        self.sock_sendto(sock_class, sock_out, peer_address)
        self.packets_transmitted += 1
        sock_class.packets_transmitted += 1
        self.pings_transmitted += 1
        sock_class.pings_transmitted += 1
        peer_state.ping_position += 1
        peer_state.sent_messages[nunpack(packet_out.message_id)] = (
            now,
            packet_out.message_id,
            peer_state.ping_position
        )
        packet_out_examine = packets.Packet()
        packet_out_examine.load(dump_out)
        if self.args.quiet:
            pass
        elif self.args.flood:
            self.print_out('.', end='', flush=True)
        if self.args.verbose:
            if self.args.encrypt:
                self.print_out('SEND (encrypted): {}'.format(repr(packet_out_examine)))
            else:
                self.print_out('SEND: {}'.format(repr(packet_out_examine)))

    def update_rtts(self, sock_class, rtt):
        for c in (self, sock_class):
            c.rtt_total += rtt
            c.rtt_total_sq += (rtt ** 2)
            c.rtt_count += 1
            if (rtt < c.rtt_min) or (c.rtt_min == 0):
                c.rtt_min = rtt
            if rtt > c.rtt_max:
                c.rtt_max = rtt
            if c.rtt_ewma == 0:
                c.rtt_ewma = rtt * 8.0
            else:
                c.rtt_ewma += (rtt - (c.rtt_ewma / 8.0))

    def sigquit_handler(self, signum, frame):
        self.print_stats(short=True)

    def sighup_handler(self, signum, frame):
        self.print_debug('Received SIGHUP, scheduling reload')
        self.is_reload = True

    def stats_time(self, seconds):
        conversion = (
            (1000, 'ms'),
            (60, 's'),
            (60, 'm'),
            (24, 'h'),
            (365, 'd'),
            (None, 'y'),
        )
        out = ''
        rest = int(seconds * 1000)
        for (div, suffix) in conversion:
            if div is None:
                if(out):
                    out = ' ' + out
                out = '{}{}{}'.format(rest, suffix, out)
                break
            p = rest % div
            rest = int(rest / div)
            if p > 0:
                if(out):
                    out = ' ' + out
                out = '{}{}{}'.format(p, suffix, out)
            if rest == 0:
                break
        return out

    def print_stats(self, short=False):
        time_end = clock()
        if self.args.listen:
            self.print_stats_sock(time_end, short=short, sock_class=None)
        else:
            for sock_class in self.sock_classes:
                self.print_stats_sock(time_end, short=short, sock_class=sock_class)

    def print_stats_sock(self, time_end, short=False, sock_class=None):
        if sock_class is not None:
            stats_class = sock_class
        else:
            stats_class = self
        time_start = self.time_start
        pings_lost = stats_class.pings_transmitted - stats_class.pings_received
        lost_pct = lazy_div(pings_lost, stats_class.pings_transmitted) * 100
        lost_undetermined = pings_lost - (stats_class.lost_outbound + stats_class.lost_inbound)
        outbound_pct = lazy_div(stats_class.lost_outbound, stats_class.pings_transmitted) * 100
        inbound_pct = lazy_div(stats_class.lost_inbound, stats_class.pings_transmitted) * 100
        undetermined_pct = lazy_div(lost_undetermined, stats_class.pings_transmitted) * 100
        rtt_avg = lazy_div(float(stats_class.rtt_total), stats_class.rtt_count)
        rtt_ewma = stats_class.rtt_ewma / 8.0
        rtt_mdev = math.sqrt(
            lazy_div(stats_class.rtt_total_sq, stats_class.rtt_count) -
            (lazy_div(stats_class.rtt_total, stats_class.rtt_count) ** 2)
        )
        if self.args.listen:
            hostname = _('Listener')
        else:
            hostname = sock_class.client_host[3]
        if short:
            self.print_out('\x0d', end='', flush=True, file=sys.stderr)
            self.print_out(_pl(
                '{hostname}: {transmitted}/{received} ping, {loss}% loss ' +
                '({outbound}/{inbound}/{undetermined} out/in/undet), min/avg/ewma/max/mdev = ' +
                '{min:0.03f}/{avg:0.03f}/{ewma:0.03f}/{max:0.03f}/{mdev:0.03f} ms',
                '{hostname}: {transmitted}/{received} pings, {loss}% loss ' +
                '({outbound}/{inbound}/{undetermined} out/in/undet), min/avg/ewma/max/mdev = ' +
                '{min:0.03f}/{avg:0.03f}/{ewma:0.03f}/{max:0.03f}/{mdev:0.03f} ms',
                stats_class.pings_received
            ).format(
                hostname=hostname,
                transmitted=stats_class.pings_transmitted,
                received=stats_class.pings_received,
                loss=int(lost_pct),
                outbound=stats_class.lost_outbound,
                inbound=stats_class.lost_inbound,
                undetermined=lost_undetermined,
                min=stats_class.rtt_min,
                avg=rtt_avg,
                ewma=rtt_ewma,
                max=stats_class.rtt_max,
                mdev=rtt_mdev,
            ), file=sys.stderr)
        elif self.args.nagios:
            if (lost_pct >= self.args.nagios_crit_loss) or (rtt_avg >= self.args.nagios_crit_rta):
                self.nagios_result = 2
                nagios_result_text = 'CRITICAL'
            elif (lost_pct >= self.args.nagios_warn_loss) or (rtt_avg >= self.args.nagios_warn_rta):
                self.nagios_result = 1
                nagios_result_text = 'WARNING'
            else:
                self.nagios_result = 0
                nagios_result_text = 'OK'
            self.print_out(_(
                '2PING {result} - Packet loss = {loss}%, RTA = {avg:0.03f} ms'
            ).format(
                result=nagios_result_text,
                loss=int(lost_pct),
                avg=rtt_avg,
            ) + (
                '|rta={avg:0.06f}ms;{avgwarn:0.06f};{avgcrit:0.06f};0.000000 ' +
                'pl={loss}%;{losswarn};{losscrit};0'
            ).format(
                avg=rtt_avg,
                loss=int(lost_pct),
                avgwarn=self.args.nagios_warn_rta,
                avgcrit=self.args.nagios_crit_rta,
                losswarn=int(self.args.nagios_warn_loss),
                losscrit=int(self.args.nagios_crit_loss),
            ))
        else:
            self.print_out('')
            self.print_out('--- {} ---'.format(_('{hostname} 2ping statistics').format(hostname=hostname)))
            self.print_out(_pl(
                '{transmitted} ping transmitted, {received} received, {loss}% ping loss, time {time}',
                '{transmitted} pings transmitted, {received} received, {loss}% ping loss, time {time}',
                stats_class.pings_transmitted
            ).format(
                transmitted=stats_class.pings_transmitted,
                received=stats_class.pings_received,
                loss=int(lost_pct),
                time=self.stats_time(time_end - time_start),
            ))
            self.print_out(_pl(
                '{outbound} outbound ping loss ({outboundpct}%), {inbound} inbound ({inboundpct}%), ' +
                '{undetermined} undetermined ({undeterminedpct}%)',
                '{outbound} outbound ping losses ({outboundpct}%), {inbound} inbound ({inboundpct}%), ' +
                '{undetermined} undetermined ({undeterminedpct}%)',
                stats_class.lost_outbound
            ).format(
                outbound=stats_class.lost_outbound,
                outboundpct=int(outbound_pct),
                inbound=stats_class.lost_inbound,
                inboundpct=int(inbound_pct),
                undetermined=lost_undetermined,
                undeterminedpct=int(undetermined_pct),
            ))
            self.print_out(_('rtt min/avg/ewma/max/mdev = {min:0.03f}/{avg:0.03f}/' +
                             '{ewma:0.03f}/{max:0.03f}/{mdev:0.03f} ms').format(
                min=stats_class.rtt_min,
                avg=rtt_avg,
                ewma=rtt_ewma,
                max=stats_class.rtt_max,
                mdev=rtt_mdev,
            ))
            self.print_out(_pl(
                '{transmitted} raw packet transmitted, {received} received',
                '{transmitted} raw packets transmitted, {received} received',
                stats_class.packets_transmitted
            ).format(
                transmitted=stats_class.packets_transmitted,
                received=stats_class.packets_received,
            ))

    def run(self):
        self.print_debug('Clock: {}, value: {:f}'.format(clock_info, clock()))
        self.print_debug('Poller: {}'.format(self.poller.poller_type))
        if hasattr(signal, 'SIGQUIT'):
            signal.signal(signal.SIGQUIT, self.sigquit_handler)
        if hasattr(signal, 'SIGHUP'):
            signal.signal(signal.SIGHUP, self.sighup_handler)

        try:
            if self.args.listen:
                self.setup_listener()
            else:
                self.setup_client()
        except (socket.error, socket.gaierror) as e:
            self.print_out(str(e))
            return 1

        try:
            self.loop()
        except KeyboardInterrupt:
            self.shutdown()
            return

    def base_packet(self):
        packet_out = packets.Packet()
        if (not self.args.no_send_version) or (self.args.notice):
            packet_out.opcodes[packets.OpcodeExtended.id] = packets.OpcodeExtended()
        if not self.args.no_send_version:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedVersion.id] = packets.ExtendedVersion()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedVersion.id].text = version_string
        if self.args.send_time:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedWallClock.id] = packets.ExtendedWallClock()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedWallClock.id].time_us = int(time.time() * 1000000)
        if self.args.send_monotonic_clock:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedMonotonicClock.id] = packets.ExtendedMonotonicClock()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedMonotonicClock.id].generation = self.fake_time_generation
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedMonotonicClock.id].time_us = int((clock() - self.time_start + self.fake_time_epoch) * 1000000)
        if self.args.send_random:
            random_data = bytes([random_sys.randint(0, 255) for x in range(self.args.send_random)])
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedRandom.id] = packets.ExtendedRandom()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedRandom.id].is_hwrng = False
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedRandom.id].is_os = has_sysrandom
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedRandom.id].random_data = random_data
        if self.args.notice:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedNotice.id] = packets.ExtendedNotice()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[packets.ExtendedNotice.id].text = self.args.notice
        if self.args.auth:
            packet_out.opcodes[packets.OpcodeHMAC.id] = packets.OpcodeHMAC()
            packet_out.opcodes[packets.OpcodeHMAC.id].key = self.args.auth.encode('UTF-8')
            packet_out.opcodes[packets.OpcodeHMAC.id].digest_index = self.args.auth_digest_index
        packet_out.padding_pattern = self.args.pattern_bytes
        packet_out.min_length = self.args.min_packet_size
        packet_out.max_length = self.args.max_packet_size
        if self.args.encrypt:
            packet_out.align_length = 16
        return packet_out

    def scheduled_cleanup(self):
        self.print_debug('Cleanup')
        for sock_class in self.sock_classes:
            self.scheduled_cleanup_sock_class(sock_class)

    def scheduled_cleanup_sock_class(self, sock_class):
        now = clock()
        for peer_tuple in tuple(sock_class.peer_states.keys()):
            peer_state = sock_class.peer_states[peer_tuple]
            if now > peer_state.last_seen + 600.0:
                del(sock_class.peer_states[peer_tuple])
                self.print_debug('Cleanup: Removed {}'.format(repr(peer_tuple)))
                continue
            for table_name, max_time in (
                ('sent_messages', 600.0),
                ('seen_messages', 600.0),
                ('courtesy_messages', 120.0),
                ('encrypted_session_ivs', 600.0),
            ):
                table = getattr(peer_state, table_name)
                for table_key_name in tuple(table.keys()):
                    if now > (table[table_key_name][0] + max_time):
                        del(table[table_key_name])
                        self.print_debug('Cleanup: Removed {} {} {}'.format(repr(peer_tuple), table_name, table_key_name))

    def new_socket(self, family, type, bind):
        sock = socket.socket(family, type)
        try:
            import IN
            sock.setsockopt(socket.IPPROTO_IP, IN.IP_RECVERR, int(True))
        except (ImportError, AttributeError, socket.error):
            pass
        if family == socket.AF_INET6:
            try:
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, int(True))
            except (AttributeError, socket.error):
                pass
        sock.bind(bind)
        self.print_debug('Bound to: {}'.format(repr((family, type, bind))))
        return sock

    def loop(self):
        while True:
            now = clock()
            if now >= self.next_cleanup:
                self.scheduled_cleanup()
                self.next_cleanup = now + 60.0
            if not self.args.listen:
                for sock_class in self.sock_classes:
                    if sock_class.shutdown_time:
                        # Scheduled to shutdown, do not send any more pings
                        continue
                    if now >= sock_class.next_send:
                        if self.args.count and (sock_class.pings_transmitted >= self.args.count):
                            sock_class.shutdown_time = now + self.args.interval
                            sock_class.next_send = now + self.args.interval
                            continue
                        if (sock_class.pings_transmitted == 0) and (self.args.preload > 1):
                            for i in range(self.args.preload):
                                self.send_new_ping(sock_class, sock_class.client_host[4])
                        else:
                            self.send_new_ping(sock_class, sock_class.client_host[4])
                        if self.args.adaptive and sock_class.rtt_ewma:
                            sock_class.next_send = now + (sock_class.rtt_ewma / 8.0 / 1000.0)
                        elif self.args.flood:
                            sock_class.next_send = now + 0.01
                        else:
                            sock_class.next_send = now + self.args.interval

            next_wakeup = now + self.old_age_interval
            next_wakeup_reason = 'old age'
            for sock_class in self.sock_classes:
                if (not self.args.listen) and (sock_class.next_send < next_wakeup):
                    next_wakeup = sock_class.next_send
                    next_wakeup_reason = 'send'
                if sock_class.shutdown_time and (sock_class.shutdown_time < next_wakeup):
                    next_wakeup = sock_class.shutdown_time
                    next_wakeup_reason = 'shutdown'
            if self.args.stats:
                if now >= self.next_stats:
                    self.print_stats(short=True)
                    self.next_stats = now + self.args.stats
                if self.next_stats < next_wakeup:
                    next_wakeup = self.next_stats
                    next_wakeup_reason = 'stats'
            if self.args.deadline:
                time_deadline = self.time_start + self.args.deadline
                if now >= time_deadline:
                    self.shutdown()
                if time_deadline < next_wakeup:
                    next_wakeup = time_deadline
                    next_wakeup_reason = 'deadline'
            if self.next_cleanup < next_wakeup:
                next_wakeup = self.next_cleanup
                next_wakeup_reason = 'cleanup'

            if next_wakeup < now:
                next_wakeup = now
                next_wakeup_reason = 'time travel'
            self.print_debug('Next wakeup: {} ({})'.format((next_wakeup - now), next_wakeup_reason))

            for sock_class in self.poller.poll(next_wakeup - now):
                try:
                    self.process_incoming_packet(sock_class)
                except Exception as e:
                    self.print_out(_('Exception: {error}').format(error=str(e)))
                    if self.args.debug:
                        raise

                if (
                    self.args.count and
                    (sock_class.pings_transmitted >= self.args.count) and
                    (sock_class.pings_transmitted == sock_class.pings_received)
                ):
                    sock_class.shutdown_time = now

            all_shutdown = True
            for sock_class in self.sock_classes:
                if (not sock_class.shutdown_time) or (sock_class.shutdown_time >= now):
                    all_shutdown = False
                    break
            if all_shutdown:
                self.shutdown()

            if self.is_reload:
                self.is_reload = False
                if self.args.listen:
                    self.setup_listener()


def main():
    args = parse_args()
    t = TwoPing(args)
    return(t.run())


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import select
import errno


class EpollPoller():
    poller_type = 'epoll'

    def __init__(self):
        self.poller = select.epoll()
        self.f_dict = {}

    def register(self, f):
        fileno = f.fileno()
        if fileno not in self.f_dict:
            self.poller.register(fileno, select.EPOLLIN)
        self.f_dict[fileno] = f

    def unregister(self, f):
        fileno = f.fileno()
        if fileno not in self.f_dict:
            return
        self.poller.unregister(fileno)
        del(self.f_dict[fileno])

    def close(self):
        return self.poller.close()

    def poll(self, timeout):
        try:
            poll_res = self.poller.poll(timeout)
        except (select.error, IOError, OSError) as e:
            if e.args[0] not in (errno.EINTR,):
                raise
            return []
        res = []
        for i in poll_res:
            if i[0] in self.f_dict:
                res.append(self.f_dict[i[0]])
        return res


class KqueuePoller():
    poller_type = 'kqueue'

    def __init__(self):
        self.poller = select.kqueue()
        self.kevents = {}
        self.f_dict = {}

    def register(self, f):
        fileno = f.fileno()
        if fileno not in self.f_dict:
            self.kevents[fileno] = select.kevent(
                fileno,
                filter=select.KQ_FILTER_READ,
                flags=select.KQ_EV_ADD | select.KQ_EV_ENABLE,
            )
        self.f_dict[fileno] = f

    def unregister(self, f):
        fileno = f.fileno()
        if fileno not in self.f_dict:
            return
        del(self.kevents[fileno])
        del(self.f_dict[fileno])

    def close(self):
        return self.poller.close()

    def poll(self, timeout):
        try:
            poll_res = self.poller.control(self.kevents.values(), 10, timeout)
        except (select.error, IOError, OSError) as e:
            if e.args[0] not in (errno.EINTR,):
                raise
            return []
        res = []
        for i in poll_res:
            if i.ident in self.f_dict:
                res.append(self.f_dict[i.ident])
        return res


class PollPoller():
    poller_type = 'poll'

    def __init__(self):
        self.poller = select.poll()
        self.f_dict = {}

    def register(self, f):
        fileno = f.fileno()
        if fileno not in self.f_dict:
            self.poller.register(fileno, select.POLLIN)
        self.f_dict[fileno] = f

    def unregister(self, f):
        fileno = f.fileno()
        if fileno not in self.f_dict:
            return
        self.poller.unregister(fileno)
        del(self.f_dict[fileno])

    def close(self):
        return self.poller.close()

    def poll(self, timeout):
        try:
            poll_res = self.poller.poll(timeout * 1000.0)
        except (select.error, IOError, OSError) as e:
            if e.args[0] not in (errno.EINTR,):
                raise
            return []
        res = []
        for i in poll_res:
            if i[0] in self.f_dict:
                res.append(self.f_dict[i[0]])
        return res


class SelectPoller():
    poller_type = 'select'

    def __init__(self):
        self.f_dict = {}

    def register(self, f):
        self.f_dict[f.fileno()] = f

    def unregister(self, f):
        fileno = f.fileno()
        if fileno not in self.f_dict:
            return
        del(self.f_dict[fileno])

    def close(self):
        pass

    def poll(self, timeout):
        try:
            return select.select(
                self.f_dict.values(),
                [],
                [],
                timeout
            )[0]
        except (select.error, IOError, OSError) as e:
            if e.args[0] not in (errno.EINTR,):
                raise
            return []


def best_poller():
    try:
        return EpollPoller()
    except AttributeError:
        pass
    try:
        return KqueuePoller()
    except AttributeError:
        pass
    try:
        return PollPoller()
    except AttributeError:
        pass
    return SelectPoller()


def available_pollers():
    available = []
    for poller in [
        EpollPoller, KqueuePoller, PollPoller, SelectPoller
    ]:
        try:
            available.append(poller())
        except AttributeError:
            continue
    return available


if __name__ == '__main__':
    available = available_pollers()
    print('Available pollers: {}'.format(' '.join([p.poller_type for p in available])))
    poller = best_poller()
    print('Best poller: {}'.format(poller.poller_type))



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import random
import hmac
import time
from . import crc32
from .utils import twoping_checksum, npack, nunpack
from math import ceil
import hashlib

try:
    from Crypto.Cipher import AES
    has_aes = True
except ImportError:
    has_aes = False


class Extended():
    id = None

    def __init__(self):
        self.data = b''

    def __repr__(self):
        if self.id is None:
            return '<Extended: {} bytes>'.format(len(self.data))
        else:
            id_hex = ''.join(['{:02x}'.format(x) for x in npack(self.id, 4)])
            return '<Extended (0x{}): {} bytes>'.format(id_hex, len(self.data))

    def load(self, data):
        self.data = data

    def dump(self):
        return self.data


class ExtendedText(Extended):
    def __init__(self):
        self.text = str()

    def __repr__(self):
        return '<Text (Generic): {}>'.format(self.text)

    def load(self, data):
        self.text = data.decode('UTF-8')

    def dump(self, max_length=None):
        text_bytes = self.text.encode('UTF-8')
        if (max_length is not None) and (max_length < len(text_bytes)):
            return None
        return text_bytes


class ExtendedVersion(ExtendedText):
    id = 0x3250564e

    def __repr__(self):
        return '<Version: {}>'.format(self.text)


class ExtendedNotice(ExtendedText):
    id = 0xa837b44e

    def __repr__(self):
        return '<Notice: {}>'.format(self.text)


class ExtendedWallClock(Extended):
    id = 0x64f69319

    def __init__(self):
        self.time_us = 0

    def __repr__(self):
        return '<Wall Clock: {}>'.format(time.strftime('%c', time.gmtime(self.time_us / 1000000.0)))

    def load(self, data):
        self.time_us = nunpack(data[0:8])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 8):
            return None
        return npack(self.time_us, 8)


class ExtendedMonotonicClock(Extended):
    id = 0x771d8dfb

    def __init__(self):
        self.generation = 0
        self.time_us = 0

    def __repr__(self):
        return '<Monotonic Clock: {:0.9f}, gen {}>'.format((self.time_us / 1000000.0), self.generation)

    def load(self, data):
        self.generation = nunpack(data[0:2])
        self.time_us = nunpack(data[2:10])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 10):
            return None
        return npack(self.generation, 2) + npack(self.time_us, 8)


class ExtendedRandom(Extended):
    id = 0x2ff6ad68

    def __init__(self):
        self.is_hwrng = False
        self.is_os = False
        self.random_data = b''

    def __repr__(self):
        return '<Random: {} ({}), HWRNG {}, OS {}>'.format(
            repr(self.random_data),
            len(self.random_data),
            repr(self.is_hwrng),
            repr(self.is_os),
        )

    def load(self, data):
        flags = nunpack(data[0:2])
        self.is_hwrng = bool(flags & 0x0001)
        self.is_os = bool(flags & 0x0002)
        self.random_data = data[2:]

    def dump(self, max_length=None):
        random_data = self.random_data
        if len(random_data) == 0:
            return None
        if max_length is not None:
            if max_length < 3:
                return None
            if max_length < (len(random_data) - 2):
                random_data = random_data[0:max_length-2]
        flags = 0
        if self.is_hwrng:
            flags = flags | 0x0001
        if self.is_os:
            flags = flags | 0x0002
        return npack(flags, 2) + random_data


class ExtendedBatteryLevels(Extended):
    id = 0x88a1f7c7

    def __init__(self):
        self.batteries = {}

    def __repr__(self):
        return '<Batteries ({}): [{}]>'.format(
            len(self.batteries),
            ', '.join(
                ['{}: {:0.03%}'.format(x, self.batteries[x] / 65535.0) for x in sorted(self.batteries)]
            ),
        )

    def load(self, data):
        self.batteries = {}
        pos = 2
        for i in range(nunpack(data[0:2])):
            battery_id = nunpack(data[pos:pos+2])
            battery_level = nunpack(data[pos+2:pos+4])
            self.batteries[battery_id] = battery_level
            pos += 4

    def dump(self, max_length=None):
        if (max_length is not None):
            if (max_length < 6):
                return None
            batteries = {}
            for i in sorted(self.batteries.keys())[0:int((max_length - 2) / 4)]:
                batteries[i] = self.batteries[i]
        else:
            batteries = self.batteries

        out = npack(len(batteries), 2)
        for i in batteries:
            out += npack(i, 2)
            out += npack(batteries[i], 2)
        return out


class Opcode():
    id = None

    def __init__(self):
        self.data = b''

    def __repr__(self):
        if self.id is None:
            return '<Opcode: {} bytes>'.format(len(self.data))
        else:
            id_hex = ''.join(['{:02x}'.format(x) for x in npack(self.id, 2)])
            return '<Opcode (0x{}): {} bytes>'.format(id_hex, len(self.data))

    def load(self, data):
        self.data = data

    def dump(self):
        return self.data


class OpcodeReplyRequested(Opcode):
    id = 0x0001

    def __init__(self):
        pass

    def __repr__(self):
        return '<Reply Requested>'

    def load(self, data):
        pass

    def dump(self, max_length=None):
        return b''


class OpcodeInReplyTo(Opcode):
    id = 0x0002

    def __init__(self):
        self.message_id = b''

    def __repr__(self):
        message_id_hex = ''.join(['{:02x}'.format(x) for x in self.message_id])
        return '<In Reply To: 0x{}>'.format(message_id_hex)

    def load(self, data):
        self.message_id = data[0:6]

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 6):
            return None
        return self.message_id


class OpcodeRTTEnclosed(Opcode):
    id = 0x0004

    def __init__(self):
        self.rtt_us = 0

    def __repr__(self):
        return '<RTT Enclosed: {} us>'.format(self.rtt_us)

    def load(self, data):
        self.rtt_us = nunpack(data[0:4])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 4):
            return None
        return npack(self.rtt_us, 4)


class OpcodeMessageIDList(Opcode):
    def __init__(self):
        self.message_ids = []

    def __repr__(self):
        ids = ['0x{}'.format(''.join(['{:02x}'.format(x) for x in y]) for y in self.message_ids)]
        return '<ID List (Generic): [{}] ({})>'.format(', '.join(ids), len(self.message_ids))

    def load(self, data):
        self.message_ids = []
        pos = 2
        for i in range(nunpack(data[0:2])):
            self.message_ids.append(data[pos:pos+6])
            pos += 6

    def dump(self, max_length=None):
        if (max_length is not None):
            if (max_length < 8):
                return None
            output_ids = self.message_ids[0:int((max_length - 2) / 6)]
        else:
            output_ids = self.message_ids

        out = npack(len(output_ids), 2)
        for i in output_ids:
            out += i
        return out


class OpcodeInvestigationSeen(OpcodeMessageIDList):
    id = 0x0008

    def __repr__(self):
        ids = ['0x{}'.format(''.join(['{:02x}'.format(x) for x in y])) for y in self.message_ids]
        return '<Investigation Seen: [{}] ({})>'.format(', '.join(ids), len(self.message_ids))


class OpcodeInvestigationUnseen(OpcodeMessageIDList):
    id = 0x0010

    def __repr__(self):
        ids = ['0x{}'.format(''.join(['{:02x}'.format(x) for x in y])) for y in self.message_ids]
        return '<Investigation Unseen: [{}] ({})>'.format(', '.join(ids), len(self.message_ids))


class OpcodeInvestigate(OpcodeMessageIDList):
    id = 0x0020

    def __repr__(self):
        ids = ['0x{}'.format(''.join(['{:02x}'.format(x) for x in y])) for y in self.message_ids]
        return '<Investigate: [{}] ({})>'.format(', '.join(ids), len(self.message_ids))


class OpcodeCourtesyExpiration(OpcodeMessageIDList):
    id = 0x0040

    def __repr__(self):
        ids = ['0x{}'.format(''.join(['{:02x}'.format(x) for x in y])) for y in self.message_ids]
        return '<Courtesy Expiration: [{}] ({})>'.format(', '.join(ids), len(self.message_ids))


class OpcodeHMAC(Opcode):
    id = 0x0080

    def __init__(self):
        self.key = b''
        self.digest_index = None
        self.hash = b''

        self.digest_map = {
            1: (hashlib.md5, 16, 'HMAC-MD5'),
            2: (hashlib.sha1, 20, 'HMAC-SHA1'),
            3: (hashlib.sha256, 32, 'HMAC-SHA256'),
            4: (crc32, 4, 'HMAC-CRC32'),
            5: (hashlib.sha512, 64, 'HMAC-SHA512'),
        }

    def __repr__(self):
        if self.digest_index is not None:
            return '<{}: 0x{}>'.format(
                self.digest_map[self.digest_index][2],
                ''.join(['{:02x}'.format(x) for x in self.hash]),
            )
        return '<HMAC>'

    def load(self, data):
        self.digest_index = nunpack(data[0:2])
        self.hash = data[2:]

    def dump(self, max_length=None):
        if self.digest_index is not None:
            (hasher, size, hasher_name) = self.digest_map[self.digest_index]
            return npack(self.digest_index, 2) + bytes(size)
        return None


class OpcodeHostLatency(Opcode):
    id = 0x0100

    def __init__(self):
        self.delay_us = 0

    def __repr__(self):
        return '<Host Latency: {} us>'.format(self.delay_us)

    def load(self, data):
        self.delay_us = nunpack(data[0:4])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 4):
            return None
        return npack(self.delay_us, 4)


class OpcodeEncrypted(Opcode):
    id = 0x0200

    def __init__(self):
        self.hkdf_info = b'\xd8\x89\xac\x93\xac\xeb\xa1\xf3\x98\xd0\xc6\x9b\xc8\xc6\xa7\xaa'
        self.method_index = None
        self.encrypted = b''
        self.session = b''
        self.iv = None

        self.method_map = {
            1: ('HKDF-AES256-CBC',),
        }

    def __repr__(self):
        if self.method_index is not None:
            return '<{} (Session {}, IV {}, {} bytes)>'.format(
                self.method_map[self.method_index][0],
                repr(self.session),
                repr(self.iv),
                len(self.encrypted),
            )
        return '<Encrypted>'

    def load(self, data):
        self.method_index = nunpack(data[0:2])
        if self.method_index == 1:
            self.session = data[2:10]
            self.iv = data[10:26]
            self.encrypted = data[26:]
        else:
            self.encrypted = data[2:]

    def dump(self, max_length=None):
        if self.method_index == 1:
            return npack(self.method_index, 2) + self.session + self.iv + self.encrypted
        return None

    def encrypt(self, unencrypted, key):
        if not has_aes:
            return None
        if self.method_index is None:
            return None
        if self.method_index == 1:
            if self.iv is None:
                self.iv = bytes([random.randint(0, 255) for x in range(16)])
            aeskey = self.hkdf(32, key, salt=self.iv, info=(self.hkdf_info + self.session), digestmod=hashlib.sha256)
            aes_e = AES.new(aeskey, AES.MODE_CBC, self.iv)
            self.encrypted = aes_e.encrypt(unencrypted)
        else:
            return None

    def decrypt(self, key):
        if not has_aes:
            return None
        if self.method_index is None:
            return None
        if self.method_index == 1:
            aeskey = self.hkdf(32, key, salt=self.iv, info=(self.hkdf_info + self.session), digestmod=hashlib.sha256)
            aes_d = AES.new(aeskey, AES.MODE_CBC, self.iv)
            return aes_d.decrypt(self.encrypted)

    def hkdf(self, length, ikm, salt=b'', info=b'', digestmod=None):
        if digestmod is None:
            digestmod = hashlib.sha256
        prk = hmac.new(salt, ikm, digestmod).digest()
        hash_len = len(prk)
        t = b''
        okm = b''
        for i in range(ceil(length / hash_len)):
            t = hmac.new(prk, t + info + bytes([1+i]), digestmod).digest()
            okm += t
        return okm[:length]


class OpcodeExtended(Opcode):
    id = 0x8000

    def __init__(self):
        self.segments = {}
        self.segment_data_positions = {}

    def __repr__(self):
        return '<Extended: {}>'.format(repr(sorted(self.segments.values(), key=lambda x: x.id)))

    def load(self, data):
        self.segments = {}
        self.segment_data_positions = {}

        pos = 0
        known_segments = (
            ExtendedVersion,
            ExtendedNotice,
            ExtendedMonotonicClock,
            ExtendedWallClock,
            ExtendedRandom,
            ExtendedBatteryLevels,
        )

        while pos < len(data):
            flag = nunpack(data[pos:pos+4])
            pos += 4
            segment_data_length = nunpack(data[pos:pos+2])
            pos += 2
            self.segment_data_positions[flag] = (pos, segment_data_length)
            segment_handler = None
            for seg in known_segments:
                if flag == seg.id:
                    segment_handler = seg
                    break
            if segment_handler is None:
                segment_handler = Extended
                segment_handler.id = flag
            self.segments[flag] = segment_handler()
            self.segments[flag].load(data[pos:(pos+segment_data_length)])
            pos += segment_data_length

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 6):
            return None
        out = b''
        pos = 0
        for segment in self.segments.values():
            if max_length is None:
                segment_max_length = None
            else:
                segment_max_length = max_length - pos - 6
            segment_data = segment.dump(max_length=segment_max_length)
            if segment_data is None:
                continue
            out += npack(segment.id, 4)
            pos += 4
            out += npack(len(segment_data), 2)
            pos += 2
            out += segment_data
            pos += len(segment_data)
        if len(out) == 0:
            return None
        return out


class Packet():
    def __repr__(self):
        return '<Packet (0x{}): {}>'.format(
            ''.join(['{:02x}'.format(x) for x in self.message_id]),
            repr(sorted(self.opcodes.values(), key=lambda x: x.id))
        )

    def __init__(self):
        self.message_id = b''
        self.opcodes = {}
        self.min_length = 0
        self.max_length = 1024
        self.align_length = 0
        self.padding_pattern = b'\x00'
        self.opcode_data_positions = {}

    def load(self, data):
        magic_number = data[0:2]
        if magic_number != b'\x32\x50':
            raise Exception('Invalid magic number')
        checksum = nunpack(data[2:4])
        if checksum:
            if twoping_checksum(data[0:2] + b'\x00\x00' + data[4:]) != checksum:
                raise Exception('Invalid checksum')
        self.message_id = data[4:10]
        opcode_flags = nunpack(data[10:12])
        self.opcodes = {}

        pos = 12
        known_opcodes = (
            OpcodeReplyRequested,
            OpcodeInReplyTo,
            OpcodeRTTEnclosed,
            OpcodeInvestigationSeen,
            OpcodeInvestigationUnseen,
            OpcodeInvestigate,
            OpcodeCourtesyExpiration,
            OpcodeHMAC,
            OpcodeHostLatency,
            OpcodeEncrypted,
            OpcodeExtended,
        )
        for flag in (2 ** x for x in range(16)):
            if not opcode_flags & flag:
                continue
            opcode_data_length = nunpack(data[pos:pos+2])
            pos += 2
            self.opcode_data_positions[flag] = (pos, opcode_data_length)
            opcode_handler = None
            for oc in known_opcodes:
                if flag == oc.id:
                    opcode_handler = oc
                    break
            if opcode_handler is None:
                opcode_handler = Opcode
                opcode_handler.id = flag
            self.opcodes[flag] = opcode_handler()
            self.opcodes[flag].load(data[pos:(pos+opcode_data_length)])
            pos += opcode_data_length

    def dump(self):
        auth_pos_begin = 0
        auth_pos_end = 0
        if not self.message_id:
            self.message_id = bytes([random.randint(0, 255) for x in range(6)])
        opcode_datas = {}
        packet_length = 12
        for flag in (
            OpcodeEncrypted.id,
            OpcodeHMAC.id,
            OpcodeReplyRequested.id,
            OpcodeInReplyTo.id,
            OpcodeRTTEnclosed.id,
            OpcodeInvestigationSeen.id,
            OpcodeInvestigationUnseen.id,
            OpcodeInvestigate.id,
            OpcodeHostLatency.id,
            OpcodeCourtesyExpiration.id,
            OpcodeExtended.id,
        ):
            if flag not in self.opcodes:
                continue
            if (packet_length + 2) > self.max_length:
                break
            res = self.opcodes[flag].dump(max_length=(self.max_length - packet_length - 2))
            if res is None:
                continue
            opcode_datas[flag] = res
            res_len = len(res)
            packet_length += res_len + 2
        opcode_flags = 0
        opcode_data = b''
        packet_length = 12
        for flag in sorted(opcode_datas.keys()):
            res = opcode_datas[flag]
            res_len = len(res)
            if flag == OpcodeHMAC.id:
                auth_pos_begin = packet_length + 4
                auth_pos_end = auth_pos_begin + (res_len - 2)
            opcode_flags = opcode_flags | flag
            opcode_data += npack(res_len, 2)
            opcode_data += res
            packet_length += res_len + 2
        out = bytearray(b'\x32\x50\x00\x00' + self.message_id + npack(opcode_flags, 2) + opcode_data)
        target_length = len(out)
        if len(out) < self.min_length:
            target_length = self.min_length
        if self.align_length and (target_length % self.align_length):
            target_length += self.align_length - (target_length % self.align_length)
        if len(out) < target_length:
            target_padding = target_length - len(out)
            padding = (self.padding_pattern * int(target_padding / len(self.padding_pattern) + 1))[0:target_padding]
            out += padding
        if (OpcodeHMAC.id in self.opcodes) and auth_pos_begin:
            out[auth_pos_begin:auth_pos_end] = self.calculate_hash(self.opcodes[OpcodeHMAC.id], out)
        out[2:4] = npack(twoping_checksum(out), 2)
        return bytes(out)

    def calculate_hash(self, opcode, payload):
        (hasher, size, hasher_name) = opcode.digest_map[opcode.digest_index]
        return hmac.new(opcode.key, payload, hasher).digest()



#!/usr/bin/env python3

import unittest
from twoping import crc32
import hmac


class TestCRC32(unittest.TestCase):
    def test_crc32(self):
        c = crc32.new()
        c.update(b'Data to hash')
        self.assertEqual(c.digest(), b'\x44\x9e\x0a\x5c')

    def test_hmac(self):
        h = hmac.new(b'Secret key', b'Data to hash', crc32)
        self.assertEqual(h.digest(), b'\x3c\xe1\xb6\xb9')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import unittest
import os
import subprocess
import time
import signal
import random
import locale
from twoping import monotonic_clock

try:
    from Crypto.Cipher import AES
    has_aes = True
except ImportError:
    has_aes = False


@unittest.skipUnless(hasattr(os, 'fork'), 'CLI tests require os.fork()')
class BaseTestCLI(unittest.TestCase):
    bind_address = '127.0.0.1'
    port = None
    settle_time = 3
    child_pid = 0
    twoing_binary = '/usr/bin/2ping'
    listener_opts = []

    @classmethod
    def setUpClass(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.twoping_binary = os.path.join(os.path.split(dir_path)[0], '2ping')
        (self.child_pid, self.port) = self.fork_listener(self, self.listener_opts)

    @classmethod
    def tearDownClass(self):
        if self.child_pid:
            os.kill(self.child_pid, signal.SIGINT)

    def fork_listener(self, extra_opts=None):
        if self.port:
            port = self.port
        else:
            port = random.randint(49152, 65535)
        opts = [
            '2ping',
            '--listen',
            '--quiet',
            '--no-3way',
            '--interface-address={}'.format(self.bind_address),
            '--port={}'.format(port),
        ]
        if extra_opts:
            opts += extra_opts

        child_pid = os.fork()
        if child_pid == 0:
            devnull_f = open(os.devnull, 'r')
            os.dup2(devnull_f.fileno(), 0)
            os.dup2(devnull_f.fileno(), 1)
            os.dup2(devnull_f.fileno(), 2)
            os.execv(self.twoping_binary, opts)
        time.sleep(self.settle_time)
        return((child_pid, port))

    def run_listener_client(self, client_opts, listener_opts=None):
        if listener_opts is None:
            listener_opts = []
        client_base_opts = [
            self.twoping_binary,
            self.bind_address,
            '--port={}'.format(self.port),
            '--debug',
            '--nagios=1000,5%,1000,5%',
        ]
        if not (('--adaptive' in client_opts) or ('--flood' in client_opts)):
            client_base_opts.append('--count=1')
        if not ('--count=1' in client_opts):
            client_base_opts.append('--interval=5')
        try:
            subprocess.check_output(client_base_opts + client_opts)
        except subprocess.CalledProcessError as e:
            self.fail(e.output)


class TestCLIStandard(BaseTestCLI):
    def test_notice(self):
        self.run_listener_client(['--notice=Notice text'])

    @unittest.skipUnless(
        (
            locale.getlocale()[1] == 'UTF-8'
        ), 'UTF-8 environment required'
    )
    def test_notice_utf8(self):
        self.run_listener_client(['--notice=UTF-8 \u2603'])

    def test_random(self):
        self.run_listener_client(['--send-random=32'])

    def test_time(self):
        self.run_listener_client(['--send-time'])

    @unittest.skipUnless(monotonic_clock.get_clock_info('clock').monotonic, 'Monotonic clock required')
    def test_monotonic_clock(self):
        self.run_listener_client(['--send-monotonic-clock'])

    def test_adaptive(self):
        self.run_listener_client(['--adaptive', '--deadline=3'])

    def test_flood(self):
        self.run_listener_client(['--flood', '--deadline=3'])


class TestCLIHMACMD5(BaseTestCLI):
    listener_opts = ['--auth-digest=hmac-md5', '--auth=rBgRpBfRbF4DkwFQXncz']

    def test_hmac(self):
        self.run_listener_client(self.listener_opts)


class TestCLIHMACSHA1(BaseTestCLI):
    listener_opts = ['--auth-digest=hmac-sha1', '--auth=qnzTCJHnZXdrxRZ8JjQw']

    def test_hmac(self):
        self.run_listener_client(self.listener_opts)


class TestCLIHMACSHA256(BaseTestCLI):
    listener_opts = ['--auth-digest=hmac-sha256', '--auth=cc8G2Ssbq4WZRq7H7d5L']

    def test_hmac(self):
        self.run_listener_client(self.listener_opts)


class TestCLIHMACSHA512(BaseTestCLI):
    listener_opts = ['--auth-digest=hmac-sha512', '--auth=sjk3kqzcSV3XfHJWNstn']

    def test_hmac(self):
        self.run_listener_client(self.listener_opts)


class TestCLIHMACCRC32(BaseTestCLI):
    listener_opts = ['--auth-digest=hmac-crc32', '--auth=mc82kJwtXFlhqQSCKptQ']

    def test_hmac(self):
        self.run_listener_client(self.listener_opts)


@unittest.skipUnless(has_aes, 'PyCrypto required')
class TestCLIEncryptAES256(BaseTestCLI):
    listener_opts = ['--encrypt-method=hkdf-aes256-cbc', '--auth=S49HVbnJd3fBdDzdMVVw']

    def test_encrypt(self):
        self.run_listener_client(self.listener_opts)


if __name__ == '__main__':
    unittest.main()



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import platform
import gettext


_ = gettext.translation('2ping', fallback=True).gettext
_pl = gettext.translation('2ping', fallback=True).ngettext


def twoping_checksum(d):
    checksum = 0

    for i in range(len(d)):
        if i & 1:
            checksum += d[i]
        else:
            checksum += d[i] << 8

    checksum = ((checksum >> 16) + (checksum & 0xffff))
    checksum = ((checksum >> 16) + (checksum & 0xffff))
    checksum = ~checksum & 0xffff

    if checksum == 0:
        checksum = 0xffff

    return checksum


def lazy_div(n, d):
    if d == 0:
        return 0
    return n / d


def npack(i, minimum=1):
    out = bytearray()
    while i >= 256:
        out.insert(0, i & 0xff)
        i = i >> 8
    out.insert(0, i)
    out_len = len(out)
    if out_len < minimum:
        out = bytearray(minimum - out_len) + out
    return bytes(out)


def nunpack(b):
    out = 0
    for x in b:
        out = (out << 8) + x
    return out


def platform_info():
    out = platform.system()
    try:
        linux_distribution = platform.linux_distribution()
        if linux_distribution[0]:
            out += ' ({})'.format(linux_distribution[0])
    except:
        pass
    out += ' {}'.format(platform.machine())
    return out



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import sys
import os
import argparse
from . import __version__
from .utils import _


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv

    if argv[0].endswith('2ping6'):
        ipv6_default = True
    else:
        ipv6_default = False

    parser = argparse.ArgumentParser(
        description='2ping ({})'.format(__version__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        prog=os.path.basename(argv[0]),
    )
    parser.add_argument(
        '--version', '-V', action='version',
        version=__version__,
        help=_('report the program version'),
    )

    # Positionals
    parser.add_argument(
        'host', type=str, default=None, nargs='*',
        help=_('host to ping'),
    )

    # ping-compatible options
    parser.add_argument(
        '--audible', '-a', dest='audible', action='store_true',
        help=_('audible ping'),
    )
    parser.add_argument(
        '--adaptive', '-A', dest='adaptive', action='store_true',
        help=_('adaptive RTT ping'),
    )
    parser.add_argument(
        '--count', '-c', dest='count', type=int,
        help=_('number of pings to send'),
    )
    parser.add_argument(
        '--flood', '-f', dest='flood', action='store_true',
        help=_('flood mode'),
    )
    parser.add_argument(
        '--interval', '-i', dest='interval', type=float, default=1.0,
        help=_('seconds between pings'), metavar='SECONDS',
    )
    parser.add_argument(
        '--interface-address', '-I', dest='interface_address', type=str, action='append',
        help=_('interface bind address'), metavar='ADDRESS',
    )
    parser.add_argument(
        '--preload', '-l', dest='preload', type=int, default=1,
        help=_('number of pings to send at start'), metavar='COUNT',
    )
    parser.add_argument(
        '--pattern', '-p', dest='pattern', type=str,
        help=_('hex pattern for padding'), metavar='HEX_BYTES',
    )
    parser.add_argument(
        '--quiet', '-q', dest='quiet', action='store_true',
        help=_('quiet mode'),
    )
    parser.add_argument(
        '--packetsize-compat', '-s', dest='packetsize_compat', type=int,
        help=_('packet size (ping compatible)'), metavar='BYTES',
    )
    parser.add_argument(
        '--verbose', '-v', dest='verbose', action='store_true',
        help=_('verbose mode'),
    )
    parser.add_argument(
        '--deadline', '-w', dest='deadline', type=float,
        help=_('maximum run time'), metavar='SECONDS',
    )

    # 2ping options
    parser.add_argument(
        '--all-interfaces', action='store_true',
        help=_('listen on all possible interface IPs'),
    )
    parser.add_argument(
        '--auth', type=str,
        help=_('HMAC authentication key'), metavar='KEY',
    )
    parser.add_argument(
        '--auth-digest', type=str, default='hmac-md5',
        choices=['hmac-md5', 'hmac-sha1', 'hmac-sha256', 'hmac-crc32', 'hmac-sha512'],
        help=_('HMAC authentication digest'), metavar='DIGEST',
    )
    parser.add_argument(
        '--debug', action='store_true',
        help=_('debug mode'),
    )
    parser.add_argument(
        '--encrypt', type=str,
        help=_('Encryption key'), metavar='KEY',
    )
    parser.add_argument(
        '--encrypt-method', type=str, default='hkdf-aes256-cbc',
        choices=['hkdf-aes256-cbc'],
        help=_('Encryption method'), metavar='METHOD',
    )
    parser.add_argument(
        '--fuzz', type=float,
        help=_('incoming fuzz percentage'), metavar='PERCENT',
    )
    parser.add_argument(
        '--inquire-wait', type=float, default=10.0,
        help=_('maximum time before loss inquiries'), metavar='SECONDS',
    )
    parser.add_argument(
        '--ipv4', '-4', action='store_true',
        help=_('force IPv4'),
    )
    parser.add_argument(
        '--ipv6', '-6', action='store_true', default=ipv6_default,
        help=_('force IPv6'),
    )
    parser.add_argument(
        '--listen', action='store_true',
        help=_('listen mode'),
    )
    parser.add_argument(
        '--max-packet-size', type=int, default=512,
        help=_('maximum packet size'), metavar='BYTES',
    )
    parser.add_argument(
        '--min-packet-size', type=int, default=128,
        help=_('minimum packet size'), metavar='BYTES',
    )
    parser.add_argument(
        '--nagios', type=str,
        help=_('nagios-compatible output'), metavar='WRTA,WLOSS%,CRTA,CLOSS%',
    )
    parser.add_argument(
        '--no-3way', action='store_true',
        help=_('do not send 3-way pings'),
    )
    parser.add_argument(
        '--no-match-packet-size', action='store_true',
        help=_('do not match packet size of peer'),
    )
    parser.add_argument(
        '--no-send-version', action='store_true',
        help=_('do not send program version to peers'),
    )
    parser.add_argument(
        '--notice', type=str,
        help=_('arbitrary notice text'), metavar='TEXT',
    )
    parser.add_argument(
        '--packet-loss', type=str,
        help=_('percentage simulated packet loss'), metavar='OUT:IN',
    )
    parser.add_argument(
        '--port', type=str, default='15998',
        help=_('port to connect / bind to'),
    )
    parser.add_argument(
        '--send-monotonic-clock', action='store_true',
        help=_('send monotonic clock to peers'),
    )
    parser.add_argument(
        '--send-random', type=int,
        help=_('send random data to peers'), metavar='BYTES',
    )
    parser.add_argument(
        '--send-time', action='store_true',
        help=_('send wall clock time to peers'),
    )
    parser.add_argument(
        '--stats', type=float,
        help=_('print recurring statistics'), metavar='SECONDS',
    )
    parser.add_argument(
        '--srv', action='store_true',
        help=_('lookup SRV records in client mode'),
    )
    parser.add_argument(
        '--srv-service', type=str, default='2ping',
        help=_('service name for SRV lookups'),
    )

    # ping-compatible ignored options
    for opt in 'b|B|d|L|n|R|r|U'.split('|'):
        parser.add_argument(
            '-{}'.format(opt), action='store_true',
            dest='ignored_{}'.format(opt),
            help=argparse.SUPPRESS,
        )
    for opt in 'F|Q|S|t|T|M|W'.split('|'):
        parser.add_argument(
            '-{}'.format(opt), type=str, default=None,
            dest='ignored_{}'.format(opt),
            help=argparse.SUPPRESS,
        )

    args = parser.parse_args(args=argv[1:])

    if (not args.listen) and (not args.host):
        parser.print_help()
        parser.exit()
    if args.nagios:
        args.quiet = True
        if (not args.count) and (not args.deadline):
            args.count = 5
        nagios_opts = args.nagios.split(',')
        if len(nagios_opts) != 4:
            parser.error(_('Invalid limits'))
        (
            args.nagios_warn_rta,
            args.nagios_warn_loss,
            args.nagios_crit_rta,
            args.nagios_crit_loss,
        ) = nagios_opts
        if args.nagios_warn_loss[-1:] != '%':
            parser.error(_('Invalid limits'))
        if args.nagios_crit_loss[-1:] != '%':
            parser.error(_('Invalid limits'))
        try:
            args.nagios_warn_loss = float(args.nagios_warn_loss[:-1])
            args.nagios_crit_loss = float(args.nagios_crit_loss[:-1])
            args.nagios_warn_rta = float(args.nagios_warn_rta)
            args.nagios_crit_rta = float(args.nagios_crit_rta)
        except ValueError as e:
            parser.error(e.message)
    if args.packetsize_compat:
        args.min_packet_size = args.packetsize_compat + 8
    if args.max_packet_size < args.min_packet_size:
        parser.error(_('Maximum packet size must be at least minimum packet size'))
    if args.max_packet_size < 64:
        parser.error(_('Maximum packet size must be at least 64'))
    args.packet_loss_in = 0
    args.packet_loss_out = 0
    if args.packet_loss:
        if ':' in args.packet_loss:
            (v_out, v_in) = args.packet_loss.split(':', 1)
            try:
                args.packet_loss_out = float(v_out)
                args.packet_loss_in = float(v_in)
            except ValueError as e:
                parser.error(e.message)
        else:
            try:
                args.packet_loss_in = args.packet_loss_out = float(args.packet_loss)
            except ValueError as e:
                parser.error(_('Packet loss: {error}').format(error=e.message))
    if args.pattern:
        if len(args.pattern) & 1:
            parser.error(_('Pattern must be full bytes'))
        if len(args.pattern) > 32:
            parser.error(_('Pattern must be 16 bytes or less'))
        args.pattern_bytes = b''
        for i in range(int(len(args.pattern) / 2)):
            a = args.pattern[(i*2):(i*2+2)]
            try:
                b = int(a, 16)
            except ValueError as e:
                parser.error(_('Pattern: {error}').format(error=e.message))
            args.pattern_bytes += bytes([b])
    else:
        args.pattern_bytes = b'\x00'

    hmac_id_map = {
        'hmac-md5': 1,
        'hmac-sha1': 2,
        'hmac-sha256': 3,
        'hmac-crc32': 4,
        'hmac-sha512': 5,
    }
    args.auth_digest_index = hmac_id_map[args.auth_digest]

    if args.encrypt:
        try:
            from Crypto.Cipher import AES
        except ImportError:
            parser.error(_('Python crypto module required for encryption'))

    encrypt_id_map = {
        'hkdf-aes256-cbc': 1,
    }
    args.encrypt_method_index = encrypt_id_map[args.encrypt_method]

    if args.debug:
        args.verbose = True

    return args



#!/usr/bin/env python3

import os
import sys
from setuptools import setup

assert(sys.version_info > (3, 4))


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


setup(
    name='2ping',
    description='2ping a bi-directional ping utility',
    long_description=read('README'),
    version='4.3',
    license='GPLv2+',
    platforms=['Unix'],
    author='Ryan Finnie',
    author_email='ryan@finnie.org',
    url='https://www.finnie.org/software/2ping/',
    download_url='https://www.finnie.org/software/2ping/',
    packages=['twoping'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet',
        'Topic :: System :: Networking',
        'Topic :: Utilities',
    ],
    entry_points={
        'console_scripts': [
            '2ping = twoping.cli:main',
            '2ping6 = twoping.cli:main',
        ],
    },
    test_suite='tests',
)



#!/usr/bin/env python3

import unittest
from twoping import packets

try:
    from Crypto.Cipher import AES
    has_aes = True
except ImportError:
    has_aes = False


class TestPacketsOpcodes(unittest.TestCase):
    def test_opcode_unknown(self):
        data = b'Unknown data'
        opcode = packets.Opcode()
        opcode.load(data)
        self.assertEqual(opcode.id, None)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_reply_requested(self):
        data = b''
        opcode = packets.OpcodeReplyRequested()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0001)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_in_reply_to(self):
        data = b'\x01\x02\x03\x04\x05\x06'
        opcode = packets.OpcodeInReplyTo()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0002)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_rtt_enclosed(self):
        data = b'\x12\x34\x56\x78'
        opcode = packets.OpcodeRTTEnclosed()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0004)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_investigation_seen(self):
        data = b'\x01\x02\x03\x04\x05\x06\x11\x12\x13\x14\x15\x16'
        opcode = packets.OpcodeInvestigationSeen()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0008)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_investigation_unseen(self):
        data = b'\x01\x02\x03\x04\x05\x06\x11\x12\x13\x14\x15\x16'
        opcode = packets.OpcodeInvestigationUnseen()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0010)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_investigate(self):
        data = b'\x01\x02\x03\x04\x05\x06\x11\x12\x13\x14\x15\x16'
        opcode = packets.OpcodeInvestigate()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0020)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_courtesy_expiration(self):
        data = b'\x01\x02\x03\x04\x05\x06\x11\x12\x13\x14\x15\x16'
        opcode = packets.OpcodeCourtesyExpiration()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0040)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_hmac(self):
        data = b'\x00\x04\x01\x02\x03\x04'
        # The dump always includes a zeroed hash area
        data_out = b'\x00\x04\x00\x00\x00\x00'
        opcode = packets.OpcodeHMAC()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0080)
        self.assertEqual(opcode.dump(), data_out)

    def test_opcode_host_latency(self):
        data = b'\x12\x34\x56\x78'
        opcode = packets.OpcodeHostLatency()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0100)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_encrypted(self):
        data = b'\x00\x01\xaa\xb1\xc0\x0f\x0f\x83\xd2\xc4x\xfe\xa1\xe2\x10by\xee"Rp\xf1\x93\xee\xe98G\xea\x11\xd1\xc9\x80=\xe3'
        opcode = packets.OpcodeEncrypted()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x0200)
        self.assertEqual(opcode.dump(), data)

    def test_opcode_extended(self):
        data = b'\x32\x50\x56\x4e\x00\x12Test 2ping version'
        opcode = packets.OpcodeExtended()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x8000)
        self.assertEqual(opcode.dump(), data)

    def test_extended_unknown(self):
        data = b'Unknown extended data'
        opcode = packets.Extended()
        opcode.load(data)
        self.assertEqual(opcode.id, None)
        self.assertEqual(opcode.dump(), data)

    def test_extended_version(self):
        data = b'Test 2ping version'
        opcode = packets.ExtendedVersion()
        opcode.load(data)
        self.assertEqual(opcode.id, 0x3250564e)
        self.assertEqual(opcode.dump(), data)

    def test_extended_notice(self):
        data = b'Notice announcement'
        opcode = packets.ExtendedNotice()
        opcode.load(data)
        self.assertEqual(opcode.id, 0xa837b44e)
        self.assertEqual(opcode.dump(), data)

    def test_extended_wallclock_load(self):
        opcode = packets.ExtendedWallClock()
        opcode.load(b'\x00\x03\x63\x73\xe7\x7a\xc2\x20')
        self.assertEqual(opcode.id, 0x64f69319)
        self.assertEqual(opcode.time_us, int(953774386.102816 * 1000000))

    def test_extended_wallclock_dump(self):
        opcode = packets.ExtendedWallClock()
        opcode.time_us = int(1454187789.993266 * 1000000)
        self.assertEqual(opcode.id, 0x64f69319)
        self.assertEqual(opcode.dump(), b'\x00\x05\x2a\x93\x7a\xa8\xc5\x32')

    def test_extended_monotonicclock_load(self):
        opcode = packets.ExtendedMonotonicClock()
        opcode.load(b'\x7d\x67\x00\x03\x63\x73\xe7\x7a\xc2\x20')
        self.assertEqual(opcode.id, 0x771d8dfb)
        self.assertEqual(opcode.generation, 32103)
        self.assertEqual(opcode.time_us, int(953774386.102816 * 1000000))

    def test_extended_monotonicclock_dump(self):
        opcode = packets.ExtendedMonotonicClock()
        opcode.generation = 9311
        opcode.time_us = int(1454187789.993266 * 1000000)
        self.assertEqual(opcode.id, 0x771d8dfb)
        self.assertEqual(opcode.dump(), b'\x24\x5f\x00\x05\x2a\x93\x7a\xa8\xc5\x32')

    def test_extended_random_load(self):
        random_data = b'\xf1\xfd\xf8\x9c\xe3\x9a\x87\x14'
        opcode = packets.ExtendedRandom()
        opcode.load(b'\x00\x03' + random_data)
        self.assertEqual(opcode.id, 0x2ff6ad68)
        self.assertTrue(opcode.is_hwrng)
        self.assertTrue(opcode.is_os)
        self.assertEqual(opcode.random_data, random_data)

    def test_extended_random_dump(self):
        random_data = b'\xbc\xde\xdc\xe5\xa8\x9a\xbd\x14'
        opcode = packets.ExtendedRandom()
        opcode.is_hwrng = True
        opcode.random_data = random_data
        self.assertEqual(opcode.id, 0x2ff6ad68)
        self.assertEqual(opcode.dump(), b'\x00\x01' + random_data)

    def test_extended_batteries_load(self):
        opcode = packets.ExtendedBatteryLevels()
        opcode.load(b'\x00\x02\x00\x00\xff\xff\x00\x01\xce\xa3')
        self.assertEqual(opcode.id, 0x88a1f7c7)
        self.assertEqual(opcode.batteries[0], 65535)
        self.assertEqual(opcode.batteries[1], 52899)

    def test_extended_batteries_dump(self):
        opcode = packets.ExtendedBatteryLevels()
        opcode.batteries = {
            0: 65535,
            1: 52899,
        }
        self.assertEqual(opcode.id, 0x88a1f7c7)
        self.assertEqual(opcode.dump(), b'\x00\x02\x00\x00\xff\xff\x00\x01\xce\xa3')

    @unittest.skipUnless(has_aes, 'PyCrypto required')
    def test_encrypted_encrypt_decrypt(self):
        key = b'Secret key'
        iv = b'2\xf0J/\xb3x\xe3\xf3s+J\x8c\x02t\xca\x0e'
        minimal_packet_data = b'2P\xda\x0a\x0e\xa5[\xe2\x89\x1d\x00\x00\x00\x00\x00\x00'
        opcode = packets.OpcodeEncrypted()
        opcode.session = b'z\xe3\xcb\xdfeK\x86\x96'
        opcode.iv = iv
        opcode.method_index = 1
        opcode.encrypt(minimal_packet_data, key)
        self.assertEqual(opcode.decrypt(key), minimal_packet_data)


class TestPacketsReference(unittest.TestCase):
    ''' Test protocol reference packets

    These tests replicate the reference packets included as examples in
    the 2ping protocol specification.
    '''
    def test_reference_1a(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xae\x00\x00\x00\x00\xa0\x01\x00\x00')

    def test_reference_2a(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xad\x00\x00\x00\x00\xa0\x01\x00\x01\x00\x00')

    def test_reference_2b(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xb0\x01'
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x01'
        self.assertEqual(packet.dump(), b'\x32\x50\x7d\xa4\x00\x00\x00\x00\xb0\x01\x00\x02\x00\x06\x00\x00\x00\x00\xa0\x01')

    def test_reference_3a(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xad\x00\x00\x00\x00\xa0\x01\x00\x01\x00\x00')

    def test_reference_3b(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xb0\x01'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x01'
        self.assertEqual(packet.dump(), b'\x32\x50\x7d\xa3\x00\x00\x00\x00\xb0\x01\x00\x03\x00\x00\x00\x06\x00\x00\x00\x00\xa0\x01')

    def test_reference_3c(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x01'
        packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
        packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
        self.assertEqual(packet.dump(), b'\x32\x50\x4d\x62\x00\x00\x00\x00\xa0\x02\x00\x06\x00\x06\x00\x00\x00\x00\xb0\x01\x00\x04\x00\x00\x30\x39')

    def test_reference_4a(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xad\x00\x00\x00\x00\xa0\x01\x00\x01\x00\x00')

    def test_reference_4b(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
        packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
        self.assertEqual(packet.dump(), b'\x32\x50\x8d\x81\x00\x00\x00\x00\xa0\x02\x00\x21\x00\x00\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x01')

    def test_reference_4c(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xb0\x02'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x02'
        packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
        packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
        self.assertEqual(packet.dump(), b'\x32\x50\xdd\x8e\x00\x00\x00\x00\xb0\x02\x00\x0b\x00\x00\x00\x06\x00\x00\x00\x00\xa0\x02\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x01')

    def test_reference_4d(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x03'
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x02'
        packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
        packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
        self.assertEqual(packet.dump(), b'\x32\x50\x4d\x60\x00\x00\x00\x00\xa0\x03\x00\x06\x00\x06\x00\x00\x00\x00\xb0\x02\x00\x04\x00\x00\x30\x39')

    def test_reference_5a(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xad\x00\x00\x00\x00\xa0\x01\x00\x01\x00\x00')

    def test_reference_5b(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
        packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
        self.assertEqual(packet.dump(), b'\x32\x50\x8d\x81\x00\x00\x00\x00\xa0\x02\x00\x21\x00\x00\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x01')

    def test_reference_5c(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xb0\x01'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x02'
        packet.opcodes[packets.OpcodeInvestigationUnseen.id] = packets.OpcodeInvestigationUnseen()
        packet.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
        self.assertEqual(packet.dump(), b'\x32\x50\xdd\x87\x00\x00\x00\x00\xb0\x01\x00\x13\x00\x00\x00\x06\x00\x00\x00\x00\xa0\x02\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x01')

    def test_reference_5d(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x03'
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x01'
        packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
        packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
        self.assertEqual(packet.dump(), b'\x32\x50\x4d\x61\x00\x00\x00\x00\xa0\x03\x00\x06\x00\x06\x00\x00\x00\x00\xb0\x01\x00\x04\x00\x00\x30\x39')

    def test_reference_6a(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xad\x00\x00\x00\x00\xa0\x01\x00\x01\x00\x00')

    def test_reference_6b(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xac\x00\x00\x00\x00\xa0\x02\x00\x01\x00\x00')

    def test_reference_6c(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x03'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        self.assertEqual(packet.dump(), b'\x32\x50\x2d\xab\x00\x00\x00\x00\xa0\x03\x00\x01\x00\x00')

    def test_reference_6d(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xb0\x02'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x03'
        self.assertEqual(packet.dump(), b'\x32\x50\x7d\xa0\x00\x00\x00\x00\xb0\x02\x00\x03\x00\x00\x00\x06\x00\x00\x00\x00\xa0\x03')

    def test_reference_6e(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x04'
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x02'
        packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
        packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12823
        self.assertEqual(packet.dump(), b'\x32\x50\x4b\x81\x00\x00\x00\x00\xa0\x04\x00\x06\x00\x06\x00\x00\x00\x00\xb0\x02\x00\x04\x00\x00\x32\x17')

    def test_reference_6f(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x0a'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
        packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
        packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x02')
        self.assertEqual(packet.dump(), b'\x32\x50\xed\x6f\x00\x00\x00\x00\xa0\x0a\x00\x21\x00\x00\x00\x0e\x00\x02\x00\x00\x00\x00\xa0\x01\x00\x00\x00\x00\xa0\x02')

    def test_reference_6g(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xb0\x06'
        packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x0a'
        packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
        packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
        packet.opcodes[packets.OpcodeInvestigationUnseen.id] = packets.OpcodeInvestigationUnseen()
        packet.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x02')
        packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
        packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xb0\x02')
        self.assertEqual(packet.dump(), b'\x32\x50\x8d\x3b\x00\x00\x00\x00\xb0\x06\x00\x3b\x00\x00\x00\x06\x00\x00\x00\x00\xa0\x0a\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x01\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x02\x00\x08\x00\x01\x00\x00\x00\x00\xb0\x02')

    def test_reference_6h(self):
        packet = packets.Packet()
        packet.message_id = b'\x00\x00\x00\x00\xa0\x0b'
        packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
        packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x06'
        packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
        packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 13112
        packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
        packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(b'\x00\x00\x00\x00\xb0\x02')
        self.assertEqual(packet.dump(), b'\x32\x50\x9a\x41\x00\x00\x00\x00\xa0\x0b\x00\x0e\x00\x06\x00\x00\x00\x00\xb0\x06\x00\x04\x00\x00\x33\x38\x00\x08\x00\x01\x00\x00\x00\x00\xb0\x02')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from twoping import packets


def h(input):
    return ' '.join('{:02x}'.format(x) for x in input)


print('### Example 1')
print()
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
print('    CLIENT: {}'.format(h(packet.dump())))
print()

print('### Example 2')
print()
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xb0\x01'
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x01'
print('    SERVER: {}'.format(h(packet.dump())))
print()

print('### Example 3')
print()
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xb0\x01'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x01'
print('    SERVER: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x01'
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
print('    CLIENT: {}'.format(h(packet.dump())))
print()

print('### Example 4')
print()
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xb0\x02'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x02'
packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
print('    SERVER: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x03'
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x02'
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
print('    CLIENT: {}'.format(h(packet.dump())))
print()

print('### Example 5')
print()
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xb0\x01'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x02'
packet.opcodes[packets.OpcodeInvestigationUnseen.id] = packets.OpcodeInvestigationUnseen()
packet.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
print('    SERVER: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x03'
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x01'
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
print('    CLIENT: {}'.format(h(packet.dump())))
print()

print('### Example 6')
print()
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x01'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x02'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x03'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xb0\x02'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x03'
print('    SERVER: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x04'
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x02'
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12823
print('    CLIENT: {}'.format(h(packet.dump())))
print('    ... etc')
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x0a'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x02')
print('    CLIENT: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xb0\x06'
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xa0\x0a'
packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x01')
packet.opcodes[packets.OpcodeInvestigationUnseen.id] = packets.OpcodeInvestigationUnseen()
packet.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids.append(b'\x00\x00\x00\x00\xa0\x02')
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(b'\x00\x00\x00\x00\xb0\x02')
print('    SERVER: {}'.format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b'\x00\x00\x00\x00\xa0\x0b'
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b'\x00\x00\x00\x00\xb0\x06'
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 13112
packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(b'\x00\x00\x00\x00\xb0\x02')
print('    CLIENT: {}'.format(h(packet.dump())))
print()



#!/usr/bin/env python3

# Monotonic clock
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import sys
import time
import types


class BuiltinMonotonicCounter():
    clock = None

    def __init__(self):
        # Unix + CLOCK_MONOTONIC_RAW
        if hasattr(time, 'clock_gettime') and hasattr(time, 'CLOCK_MONOTONIC_RAW'):
            self.info = time.get_clock_info('monotonic')
            self.info.implementation = 'clock_gettime(CLOCK_MONOTONIC_RAW)'

            def _clock():
                return time.clock_gettime(time.CLOCK_MONOTONIC_RAW)

            self.clock = _clock
            self.clock()
            return

        # All other platforms
        for mode in ('perf_counter', 'monotonic', 'clock'):
            try:
                self.info = time.get_clock_info(mode)
            except ValueError:
                continue
            if not self.info.monotonic:
                continue
            if self.info.resolution > 1e-06:
                continue
            self.clock = getattr(time, mode)
            self.clock()
            return

        raise NotImplementedError()


class SystemCounter():
    info = types.SimpleNamespace(
        adjustable=True,
        implementation='time()',
        monotonic=False,
        resolution=1.0,
    )

    def clock(self):
        return time.time()


try:
    _counter = BuiltinMonotonicCounter()
except RuntimeError:
    # Fall back to time.time() as a last resort
    _counter = SystemCounter()

clock = _counter.clock
monotonic = _counter.clock


def get_clock_info(clock):
    if clock not in ('clock', 'monotonic'):
        raise ValueError('unknown clock')
    return _counter.info


if __name__ == '__main__':
    if len(sys.argv) > 1:
        for _ in range(int(sys.argv[1])):
            print('{:.50f}'.format(clock()))
    else:
        print(get_clock_info('clock'))
        previous = 0.0
        for _ in range(10):
            current = clock()
            if current >= previous:
                print('{:.50f}'.format(current))
            else:
                print('{:.50f} (!!!)'.format(current))



#!/usr/bin/env python3

import unittest
from twoping import best_poller
import socket


class TestBestPoller(unittest.TestCase):
    def test_register(self):
        poller = best_poller.best_poller()
        s = socket.socket()
        poller.register(s)
        self.assertEqual(
            poller.f_dict[s.fileno()],
            s
        )
        s.close()

    def test_register_multiple(self):
        poller = best_poller.best_poller()
        s1 = socket.socket()
        s2 = socket.socket()
        poller.register(s1)
        poller.register(s2)
        self.assertEqual(
            (poller.f_dict[s1.fileno()], poller.f_dict[s2.fileno()]),
            (s1, s2)
        )
        s1.close()
        s2.close()

    def test_register_idempotent(self):
        poller = best_poller.best_poller()
        s = socket.socket()
        poller.register(s)
        poller.register(s)
        self.assertEqual(
            len(poller.f_dict),
            1
        )
        s.close()

    def test_available_pollers(self):
        self.assertTrue(len(best_poller.available_pollers()) > 0)

    def test_known_poller(self):
        self.assertTrue(isinstance(
            best_poller.best_poller(),
            (
                best_poller.EpollPoller,
                best_poller.KqueuePoller,
                best_poller.PollPoller,
                best_poller.SelectPoller,
            )
        ))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2018 Ryan Finnie
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import binascii
import copy

digest_size = 4


class CRC32():
    digest_size = 4
    block_size = 64
    _crc = 0

    def __init__(self, buf=None):
        if buf is not None:
            self.update(buf)

    def copy(self):
        return copy.copy(self)

    def update(self, buf):
        self._crc = binascii.crc32(buf, self._crc)

    def clear(self):
        self._crc = 0

    def digest(self):
        i = self._crc & 0xffffffff
        out = bytearray()
        while i >= 256:
            out.insert(0, i & 0xff)
            i = i >> 8
        out.insert(0, i)
        out_len = len(out)
        if out_len < 4:
            out = bytearray(4 - out_len) + out
        return out

    def hexdigest(self):
        return ''.join('{hex:02x}'.format(hex=x) for x in self.digest())


def new(buf=None):
    return CRC32(buf)


if __name__ == '__main__':
    import sys
    files = sys.argv[1:]
    if len(files) == 0:
        if hasattr(sys.stdin, 'buffer'):
            stdin = sys.stdin.buffer
        else:
            stdin = sys.stdin
        c = new()
        for buf in stdin.readlines():
            c.update(buf)
        print(c.hexdigest())
    else:
        for file in files:
            with open(file, 'rb') as f:
                c = new()
                for buf in f.readlines():
                    c.update(buf)
                if len(files) > 1:
                    print('{}\t{}'.format(c.hexdigest(), file))
                else:
                    print(c.hexdigest())
